import json
from pathlib import Path

from nf_core.pipelines.params_file import ParamsFileBuilder

from ..test_pipelines import TestPipelines


class TestParamsFileBuilder(TestPipelines):
    """Class for schema tests"""

    def setUp(self):
        """Create a new PipelineSchema object"""
        super().setUp()

        self.template_schema = Path(self.pipeline_dir, "nextflow_schema.json")
        self.params_template_builder = ParamsFileBuilder(self.pipeline_dir)
        self.outfile = Path(self.pipeline_dir, "params-file.yml")

    def test_build_template(self):
        self.params_template_builder.write_params_file(self.outfile)

        assert self.outfile.exists()

        with open(self.outfile) as fh:
            out = fh.read()

        assert "nf-core/testpipeline" in out

    def test_build_template_invalid_schema(self):
        """Build a schema from a template"""
        schema = {}
        with open(self.template_schema) as fh:
            schema = json.load(fh)
            del schema["allOf"]

        with open(self.template_schema, "w") as fh:
            json.dump(schema, fh)

        builder = ParamsFileBuilder(self.template_schema)
        res = builder.write_params_file(self.outfile)

        assert res is False
        assert "Pipeline schema file is invalid" in self.caplog.text

    def test_build_template_file_exists(self):
        """Build a schema from a template"""

        # Creates a new empty file
        self.outfile.touch()

        res = self.params_template_builder.write_params_file(self.outfile)

        assert res is False
        assert f"File '{self.outfile}' exists!" in self.caplog.text

        self.outfile.unlink()

    def test_build_template_content(self):
        """Test that the content of the params file is correct"""
        self.params_template_builder.write_params_file(self.outfile)

        with open(self.outfile) as fh:
            out = fh.read()

        assert "nf-core/testpipeline" in out
        assert "# input = null" in out



"""
Code for linting modules and subworkflows in the nf-core/modules repository and
in nf-core pipelines
"""

import logging
import operator
import os
from pathlib import Path
from typing import List, Optional, Tuple, Union

import rich.box
import rich.console
import rich.panel
import rich.repr
from rich.markdown import Markdown
from rich.table import Table

import nf_core.modules.modules_utils
import nf_core.utils
from nf_core.components.components_command import ComponentCommand
from nf_core.components.nfcore_component import NFCoreComponent
from nf_core.modules.modules_json import ModulesJson
from nf_core.pipelines.lint_utils import console
from nf_core.utils import NFCoreYamlLintConfig
from nf_core.utils import plural_s as _s

log = logging.getLogger(__name__)


class LintExceptionError(Exception):
    """Exception raised when there was an error with module or subworkflow linting"""

    pass


@rich.repr.auto
class LintResult:
    """An object to hold the results of a lint test"""

    def __init__(self, component: NFCoreComponent, lint_test: str, message: str, file_path: Path):
        self.component = component
        self.lint_test = lint_test
        self.message = message
        self.file_path = file_path
        self.component_name: str = component.component_name


class ComponentLint(ComponentCommand):
    """
    An object for linting modules and subworkflows either in a clone of the 'nf-core/modules'
    repository or in any nf-core pipeline directory
    """

    def __init__(
        self,
        component_type: str,
        directory: Union[str, Path],
        fail_warned: bool = False,
        fix: bool = False,
        remote_url: Optional[str] = None,
        branch: Optional[str] = None,
        no_pull: bool = False,
        registry: Optional[str] = None,
        hide_progress: bool = False,
    ):
        super().__init__(
            component_type,
            directory=directory,
            remote_url=remote_url,
            branch=branch,
            no_pull=no_pull,
            hide_progress=hide_progress,
        )

        self.fail_warned = fail_warned
        self.fix = fix
        self.passed: List[LintResult] = []
        self.warned: List[LintResult] = []
        self.failed: List[LintResult] = []
        self.all_local_components: List[NFCoreComponent] = []

        self.lint_config: Optional[NFCoreYamlLintConfig] = None
        self.modules_json: Optional[ModulesJson] = None

        if self.component_type == "modules":
            self.lint_tests = self.get_all_module_lint_tests(self.repo_type == "pipeline")
        else:
            self.lint_tests = self.get_all_subworkflow_lint_tests(self.repo_type == "pipeline")

        if self.repo_type is None:
            raise LookupError(
                "Could not determine repository type. Please check the repository type in the nf-core.yml"
            )

        if self.repo_type == "pipeline":
            modules_json = ModulesJson(self.directory)
            modules_json.check_up_to_date()
            self.all_remote_components: List[NFCoreComponent] = []
            for repo_url, components in modules_json.get_all_components(self.component_type).items():
                if remote_url is not None and remote_url != repo_url:
                    continue
                if isinstance(components, str):
                    raise LookupError(
                        f"Error parsing modules.json: {components}. " f"Please check the file for errors or try again."
                    )
                for org, comp in components:
                    self.all_remote_components.append(
                        NFCoreComponent(
                            comp,
                            repo_url,
                            Path(self.directory, self.component_type, org, comp),
                            self.repo_type,
                            self.directory,
                            self.component_type,
                        )
                    )
            if not self.all_remote_components:
                log.warning(f"No {self.component_type} from {self.modules_repo.remote_url} installed in pipeline.")
            local_component_dir = Path(self.directory, self.component_type, "local")

            if local_component_dir.exists():
                self.all_local_components = [
                    NFCoreComponent(
                        comp,
                        None,
                        Path(local_component_dir, comp),
                        self.repo_type,
                        self.directory,
                        self.component_type,
                        remote_component=False,
                    )
                    for comp in self.get_local_components()
                ]
            self.config = nf_core.utils.fetch_wf_config(self.directory, cache_config=True)
            self._set_registry(registry)

        elif self.repo_type == "modules":
            component_dir = Path(
                self.directory,
                self.default_modules_path if self.component_type == "modules" else self.default_subworkflows_path,
            )
            self.all_remote_components = [
                NFCoreComponent(m, None, component_dir / m, self.repo_type, self.directory, self.component_type)
                for m in self.get_components_clone_modules()
            ]
            self.all_local_components = []
            if not self.all_remote_components:
                log.warning(f"No {self.component_type} in '{self.component_type}' directory")

            # This could be better, perhaps glob for all nextflow.config files in?
            self.config = nf_core.utils.fetch_wf_config(self.directory / "tests" / "config", cache_config=True)
            self._set_registry(registry)

    def __repr__(self) -> str:
        return f"ComponentLint({self.component_type}, {self.directory})"

    def _set_registry(self, registry) -> None:
        if registry is None:
            self.registry = self.config.get("docker.registry", "quay.io")
        else:
            self.registry = registry
        log.debug(f"Registry set to {self.registry}")

    @staticmethod
    def get_all_module_lint_tests(is_pipeline):
        if is_pipeline:
            return [
                "environment_yml",
                "module_patch",
                "module_version",
                "main_nf",
                "meta_yml",
                "module_todos",
                "module_deprecations",
                "module_changes",
            ]
        else:
            return ["environment_yml", "main_nf", "meta_yml", "module_todos", "module_deprecations", "module_tests"]

    @staticmethod
    def get_all_subworkflow_lint_tests(is_pipeline):
        if is_pipeline:
            return ["main_nf", "meta_yml", "subworkflow_changes", "subworkflow_todos", "subworkflow_version"]
        else:
            return ["main_nf", "meta_yml", "subworkflow_todos", "subworkflow_tests"]

    def set_up_pipeline_files(self):
        self.load_lint_config()
        self.modules_json = ModulesJson(self.directory)
        self.modules_json.load()

        # Only continue if a lint config has been loaded
        if self.lint_config:
            for test_name in self.lint_tests:
                if self.lint_config.get(test_name, {}) is False:
                    log.info(f"Ignoring lint test: {test_name}")
                    self.lint_tests.remove(test_name)

    def filter_tests_by_key(self, key):
        """Filters the tests by the supplied key"""
        # Check that supplied test keys exist
        bad_keys = [k for k in key if k not in self.lint_tests]
        if len(bad_keys) > 0:
            raise AssertionError(
                "Test name{} not recognised: '{}'".format(
                    _s(bad_keys),
                    "', '".join(bad_keys),
                )
            )

        # If -k supplied, only run these tests
        self.lint_tests = [k for k in self.lint_tests if k in key]

    def _print_results(self, show_passed=False, sort_by="test"):
        """Print linting results to the command line.

        Uses the ``rich`` library to print a set of formatted tables to the command line
        summarising the linting results.
        """

        log.debug("Printing final results")

        sort_order = ["lint_test", "component_name", "message"]
        if sort_by == "module" or sort_by == "subworkflow":
            sort_order = ["component_name", "lint_test", "message"]

        # Sort the results
        self.passed.sort(key=operator.attrgetter(*sort_order))
        self.warned.sort(key=operator.attrgetter(*sort_order))
        self.failed.sort(key=operator.attrgetter(*sort_order))

        # Find maximum module name length
        max_name_len = len(self.component_type[:-1] + " name")
        for tests in [self.passed, self.warned, self.failed]:
            try:
                for lint_result in tests:
                    max_name_len = max(len(lint_result.component_name), max_name_len)
            except Exception:
                pass

        # Helper function to format test links nicely
        def format_result(test_results, table):
            """
            Given an list of error message IDs and the message texts, return a nicely formatted
            string for the terminal with appropriate ASCII colours.
            """
            # TODO: Row styles don't work current as table-level style overrides.
            # Leaving it here in case there is a future fix
            last_modname = False
            even_row = False
            for lint_result in test_results:
                if last_modname and lint_result.component_name != last_modname:
                    even_row = not even_row
                last_modname = lint_result.component_name

                # If this is an nf-core module, link to the nf-core webpage
                if lint_result.component.repo_url == "https://github.com/nf-core/modules.git":
                    module_url = "https://nf-co.re/modules/" + lint_result.component_name.replace("/", "_")
                    module_name = f"[link={module_url}]{lint_result.component_name}[/link]"
                else:
                    module_name = lint_result.component_name

                # Make the filename clickable to open in VSCode
                file_path = os.path.relpath(lint_result.file_path, self.directory)
                file_path_link = f"[link=vscode://file/{os.path.abspath(file_path)}]{file_path}[/link]"

                table.add_row(
                    module_name,
                    file_path_link,
                    Markdown(f"{lint_result.message}"),
                    style="dim" if even_row else None,
                )
            return table

        # Print blank line for spacing
        console.print("")

        # Table of passed tests
        if len(self.passed) > 0 and show_passed:
            table = Table(style="green", box=rich.box.MINIMAL, pad_edge=False, border_style="dim")
            table.add_column(f"{self.component_type[:-1].title()} name", width=max_name_len)
            table.add_column("File path")
            table.add_column("Test message")
            table = format_result(self.passed, table)
            console.print(
                rich.panel.Panel(
                    table,
                    title=rf"[bold][✔] {len(self.passed)} {self.component_type[:-1].title()} Test{_s(self.passed)} Passed",
                    title_align="left",
                    style="green",
                    padding=0,
                )
            )

        # Table of warning tests
        if len(self.warned) > 0:
            table = Table(style="yellow", box=rich.box.MINIMAL, pad_edge=False, border_style="dim")
            table.add_column(f"{self.component_type[:-1].title()} name", width=max_name_len)
            table.add_column("File path")
            table.add_column("Test message", overflow="fold")
            table = format_result(self.warned, table)
            console.print(
                rich.panel.Panel(
                    table,
                    title=rf"[bold][!] {len(self.warned)} {self.component_type[:-1].title()} Test Warning{_s(self.warned)}",
                    title_align="left",
                    style="yellow",
                    padding=0,
                )
            )

        # Table of failing tests
        if len(self.failed) > 0:
            table = Table(
                style="red",
                box=rich.box.MINIMAL,
                pad_edge=False,
                border_style="dim",
            )
            table.add_column(f"{self.component_type[:-1].title()} name", width=max_name_len)
            table.add_column("File path")
            table.add_column("Test message", overflow="fold")
            table = format_result(self.failed, table)
            console.print(
                rich.panel.Panel(
                    table,
                    title=rf"[bold][✗] {len(self.failed)} {self.component_type[:-1].title()} Test{_s(self.failed)} Failed",
                    title_align="left",
                    style="red",
                    padding=0,
                )
            )

    def print_summary(self):
        """Print a summary table to the console."""
        table = Table(box=rich.box.ROUNDED)
        table.add_column("[bold green]LINT RESULTS SUMMARY", no_wrap=True)
        table.add_row(
            rf"[✔] {len(self.passed):>3} Test{_s(self.passed)} Passed",
            style="green",
        )
        table.add_row(rf"[!] {len(self.warned):>3} Test Warning{_s(self.warned)}", style="yellow")
        table.add_row(rf"[✗] {len(self.failed):>3} Test{_s(self.failed)} Failed", style="red")
        console.print(table)



"""
Lint the tests of a subworkflow in nf-core/modules
"""

import json
import logging
import re
from pathlib import Path

import yaml

from nf_core.components.nfcore_component import NFCoreComponent

log = logging.getLogger(__name__)


def subworkflow_tests(_, subworkflow: NFCoreComponent):
    """
    Lint the tests of a subworkflow in ``nf-core/modules``

    It verifies that the test directory exists
    and contains a ``main.nf.test`` and a ``main.nf.test.snap``

    Additionally, checks that all included components in test ``main.nf`` are specified in ``test.yml``
    """

    repo_dir = subworkflow.component_dir.parts[
        : subworkflow.component_dir.parts.index(subworkflow.component_name.split("/")[0])
    ][-1]
    pytest_dir = Path(
        subworkflow.base_dir,
        "tests",
        "subworkflows",
        repo_dir,
        subworkflow.component_name,
    )
    pytest_main_nf = Path(pytest_dir, "main.nf")
    is_pytest = pytest_main_nf.is_file()
    log.debug(f"{pytest_main_nf} is pytest: {is_pytest}")
    if subworkflow.nftest_testdir.is_dir():
        subworkflow.passed.append(
            (
                "test_dir_exists",
                "nf-test test directory exists",
                subworkflow.nftest_testdir,
            )
        )
    else:
        if is_pytest:
            subworkflow.warned.append(
                (
                    "test_dir_exists",
                    "Migrate pytest-workflow to nf-test",
                    subworkflow.nftest_testdir,
                )
            )
        else:
            subworkflow.failed.append(
                (
                    "test_dir_exists",
                    "nf-test directory is missing",
                    subworkflow.nftest_testdir,
                )
            )
        return

    # Lint the test main.nf file
    if subworkflow.nftest_main_nf.is_file():
        subworkflow.passed.append(
            (
                "test_main_nf_exists",
                "test `main.nf.test` exists",
                subworkflow.nftest_main_nf,
            )
        )
    else:
        if is_pytest:
            subworkflow.warned.append(
                (
                    "test_main_nf_exists",
                    "test `main.nf.test` does not exist",
                    subworkflow.nftest_main_nf,
                )
            )
        else:
            subworkflow.failed.append(
                (
                    "test_main_nf_exists",
                    "test `main.nf.test` does not exist",
                    subworkflow.nftest_main_nf,
                )
            )

    if subworkflow.nftest_main_nf.is_file():
        with open(subworkflow.nftest_main_nf) as fh:
            # Check if main.nf.test.snap file exists, if 'snap(' is inside main.nf.test
            if "snapshot(" in fh.read():
                snap_file = subworkflow.nftest_testdir / "main.nf.test.snap"
                if snap_file.is_file():
                    subworkflow.passed.append(
                        (
                            "test_snapshot_exists",
                            "test `main.nf.test.snap` exists",
                            snap_file,
                        )
                    )
                    # Validate no empty files
                    with open(snap_file) as snap_fh:
                        try:
                            snap_content = json.load(snap_fh)
                            for test_name in snap_content.keys():
                                if "d41d8cd98f00b204e9800998ecf8427e" in str(snap_content[test_name]):
                                    if "stub" not in test_name:
                                        subworkflow.failed.append(
                                            (
                                                "test_snap_md5sum",
                                                "md5sum for empty file found: d41d8cd98f00b204e9800998ecf8427e",
                                                snap_file,
                                            )
                                        )
                                    else:
                                        subworkflow.passed.append(
                                            (
                                                "test_snap_md5sum",
                                                "md5sum for empty file found, but it is a stub test",
                                                snap_file,
                                            )
                                        )
                                else:
                                    subworkflow.passed.append(
                                        (
                                            "test_snap_md5sum",
                                            "no md5sum for empty file found",
                                            snap_file,
                                        )
                                    )
                                if "7029066c27ac6f5ef18d660d5741979a" in str(snap_content[test_name]):
                                    if "stub" not in test_name:
                                        subworkflow.failed.append(
                                            (
                                                "test_snap_md5sum",
                                                "md5sum for compressed empty file found: 7029066c27ac6f5ef18d660d5741979a",
                                                snap_file,
                                            )
                                        )
                                    else:
                                        subworkflow.failed.append(
                                            (
                                                "test_snap_md5sum",
                                                "md5sum for compressed empty file found, but it is a stub test",
                                                snap_file,
                                            )
                                        )
                                else:
                                    subworkflow.passed.append(
                                        (
                                            "test_snap_md5sum",
                                            "no md5sum for compressed empty file found",
                                            snap_file,
                                        )
                                    )
                            if "versions" in str(snap_content[test_name]) or "versions" in str(snap_content.keys()):
                                subworkflow.passed.append(
                                    (
                                        "test_snap_versions",
                                        "versions found in snapshot file",
                                        snap_file,
                                    )
                                )
                            else:
                                subworkflow.warned.append(
                                    (
                                        "test_snap_versions",
                                        "versions not found in snapshot file",
                                        snap_file,
                                    )
                                )
                        except json.decoder.JSONDecodeError as e:
                            subworkflow.failed.append(
                                (
                                    "test_snapshot_exists",
                                    f"snapshot file `main.nf.test.snap` can't be read: {e}",
                                    snap_file,
                                )
                            )
                else:
                    subworkflow.failed.append(
                        (
                            "test_snapshot_exists",
                            "test `main.nf.test.snap` does not exist",
                            snap_file,
                        )
                    )
            # Verify that tags are correct.
            main_nf_tags = subworkflow._get_main_nf_tags(subworkflow.nftest_main_nf)
            not_alphabet = re.compile(r"[^a-zA-Z]")
            org_alp = not_alphabet.sub("", subworkflow.org)
            org_alphabet = org_alp if org_alp != "" else "nfcore"
            required_tags = [
                "subworkflows",
                f"subworkflows/{subworkflow.component_name}",
                f"subworkflows_{org_alphabet}",
            ]
            included_components = []
            if subworkflow.main_nf is not None and Path(subworkflow.main_nf).is_file():
                included_components = subworkflow._get_included_components(subworkflow.main_nf)
            if subworkflow.nftest_main_nf is not None and subworkflow.nftest_main_nf.is_file():
                chained_components_tags = subworkflow._get_included_components_in_chained_tests(
                    subworkflow.nftest_main_nf
                )
            log.debug(f"Included components: {included_components}")
            log.debug(f"Required tags: {required_tags}")
            log.debug(f"Included components for chained nf-tests: {chained_components_tags}")
            missing_tags = []
            for tag in set(required_tags + included_components + chained_components_tags):
                if tag not in main_nf_tags:
                    missing_tags.append(tag)
            if len(missing_tags) == 0:
                subworkflow.passed.append(
                    (
                        "test_main_tags",
                        "Tags adhere to guidelines",
                        subworkflow.nftest_main_nf,
                    )
                )
            else:
                subworkflow.failed.append(
                    (
                        "test_main_tags",
                        f"Tags do not adhere to guidelines. Tags missing in `main.nf.test`: {missing_tags}",
                        subworkflow.nftest_main_nf,
                    )
                )

    # Check pytest_modules.yml does not contain entries for subworkflows with nf-test
    pytest_yml_path = subworkflow.base_dir / "tests" / "config" / "pytest_modules.yml"
    if pytest_yml_path.is_file() and not is_pytest:
        try:
            with open(pytest_yml_path) as fh:
                pytest_yml = yaml.safe_load(fh)
                if "subworkflows/" + subworkflow.component_name in pytest_yml.keys():
                    subworkflow.failed.append(
                        (
                            "test_pytest_yml",
                            "subworkflow with nf-test should not be listed in pytest_modules.yml",
                            pytest_yml_path,
                        )
                    )
                else:
                    subworkflow.passed.append(
                        (
                            "test_pytest_yml",
                            "subworkflow with  nf-test not in pytest_modules.yml",
                            pytest_yml_path,
                        )
                    )
        except FileNotFoundError:
            subworkflow.warned.append(
                (
                    "test_pytest_yml",
                    "Could not open pytest_modules.yml file",
                    pytest_yml_path,
                )
            )

    # Check that the old test directory does not exist
    if not is_pytest:
        if pytest_dir.is_dir():
            subworkflow.failed.append(("test_old_test_dir", "old test directory exists", pytest_dir))
        else:
            subworkflow.passed.append(("test_old_test_dir", "old test directory does not exist", pytest_dir))



import logging

from nf_core.pipelines.lint.pipeline_todos import pipeline_todos

log = logging.getLogger(__name__)


def subworkflow_todos(_, subworkflow):
    """
    Look for TODO statements in the subworkflow files

    The nf-core subworkflow template contains a number of comment lines to help developers
    of new subworkflow know where they need to edit files and add content.
    They typically have the following format:

    .. code-block:: groovy

        // TODO nf-core: Make some kind of change to the workflow here

    ..or in markdown:

    .. code-block:: html

        <!-- TODO nf-core: Add some detail to the docs here -->

    This lint test runs through all files in the subworkflows and searches for these lines.
    If any are found they will throw a warning.

    .. tip:: Note that many GUI code editors have plugins to list all instances of *TODO*
              in a given project directory. This is a very quick and convenient way to get
              started on your pipeline!

    """

    # Main subworkflow directory
    swf_results = pipeline_todos(None, root_dir=subworkflow.component_dir)
    for i, warning in enumerate(swf_results["warned"]):
        subworkflow.warned.append(("subworkflow_todo", warning, swf_results["file_paths"][i]))
    for i, passed in enumerate(swf_results["passed"]):
        subworkflow.passed.append(("subworkflow_todo", passed, subworkflow.component_dir))



from pathlib import Path
from typing import Dict, List

import yaml


def actions_awsfulltest(self) -> Dict[str, List[str]]:
    """Checks the GitHub Actions awsfulltest is valid.

    In addition to small test datasets run on GitHub Actions, we provide the possibility of testing the pipeline on full size datasets on AWS.
    This should ensure that the pipeline runs as expected on AWS and provide a resource estimation.

    The GitHub Actions workflow is called ``awsfulltest.yml``, and it can be found in the ``.github/workflows/`` directory.

    .. warning::  This workflow incurs AWS costs, therefore it should only be triggered for pipeline releases:
                  ``release`` (after the pipeline release) and ``workflow_dispatch``.

    .. note::  You can manually trigger the AWS tests by going to the `Actions` tab on the pipeline GitHub repository and selecting the
                  `nf-core AWS full size tests` workflow on the left.

    .. tip::  For tests on full data prior to release, `Seqera Platform <https://seqera.io/platform/>`_ launch feature can be employed.

    The ``.github/workflows/awsfulltest.yml`` file is tested for the following:

    * Must be turned on ``workflow_dispatch``.
    * Must be turned on for ``release`` with ``types: [published]``
    * Should run the profile ``test_full`` that should be edited to provide the links to full-size datasets. If it runs the profile ``test``, a warning is given.
    """
    passed = []
    warned = []
    failed = []

    fn = Path(self.wf_path, ".github", "workflows", "awsfulltest.yml")
    if fn.is_file():
        try:
            with open(fn) as fh:
                wf = yaml.safe_load(fh)
        except Exception as e:
            return {"failed": [f"Could not parse yaml file: {fn}, {e}"]}

        aws_profile = "-profile test "

        # Check that the action is only turned on for published releases
        try:
            if wf[True]["pull_request"]["branches"] != ["main", "master"]:
                raise AssertionError()
            if wf[True]["pull_request_review"]["types"] != ["submitted"]:
                raise AssertionError()
            if "workflow_dispatch" not in wf[True]:
                raise AssertionError()
        except (AssertionError, KeyError, TypeError):
            failed.append("`.github/workflows/awsfulltest.yml` is not triggered correctly")
        else:
            passed.append("`.github/workflows/awsfulltest.yml` is triggered correctly")

        # Warn if `-profile test` is still unchanged
        try:
            steps = wf["jobs"]["run-platform"]["steps"]
            if not any(aws_profile in step["run"] for step in steps if "run" in step.keys()):
                raise AssertionError()
        except (AssertionError, KeyError, TypeError):
            passed.append("`.github/workflows/awsfulltest.yml` does not use `-profile test`")
        else:
            warned.append("`.github/workflows/awsfulltest.yml` should test full datasets, not `-profile test`")

    return {"passed": passed, "warned": warned, "failed": failed}



from pathlib import Path

from .utils import with_temporary_file, with_temporary_folder


def test_with_temporary_file():
    @with_temporary_file
    def tmp_file_exists(tmp_file):
        assert Path(tmp_file.name).exists()

    tmp_file_exists()


def test_does_not_exist_after():
    tmp_file = with_temporary_file(lambda x: x.name)()
    assert not Path(tmp_file).exists()


def test_with_temporary_folder():
    @with_temporary_folder
    def tmp_folder_exists(tmp_folder):
        assert Path(tmp_folder).exists()

    tmp_folder_exists()


def test_tmp_folder_does_not_exist_after():
    tmp_folder = with_temporary_folder(lambda x: x)()
    assert not Path(tmp_folder).exists()



import nf_core.pipelines.schema


def schema_description(self):
    """Check that every parameter in the schema has a description.

    The ``nextflow_schema.json`` pipeline schema should describe every flat parameter.

    Furthermore warns about parameters outside of groups.

    * Warning: Parameters in ``nextflow_schema.json`` without a description
    * Warning: Parameters in ``nextflow_schema.json`` that are defined outside of a group
    """
    passed = []
    warned = []
    ignored = []

    # First, get the top-level config options for the pipeline
    # Schema object already created in the `schema_lint` test
    self.schema_obj = nf_core.pipelines.schema.PipelineSchema()
    self.schema_obj.get_schema_path(self.wf_path)
    self.schema_obj.get_wf_params()
    self.schema_obj.no_prompts = True
    self.schema_obj.load_lint_schema()

    # Get parameters that should be ignored according to the linting config
    ignore_params = self.lint_config.get("schema_description", []) if self.lint_config is not None else []

    # Get ungrouped params
    if "properties" in self.schema_obj.schema.keys():
        ungrouped_params = self.schema_obj.schema["properties"].keys()
        for up in ungrouped_params:
            if up in ignore_params:
                ignored.append(f"Ignored ungrouped param in schema: `{up}`")
            else:
                warned.append(f"Ungrouped param in schema: `{up}`")

    # Iterate over groups and add warning for parameters without a description
    defs_notation = self.schema_obj.defs_notation
    for group_key in self.schema_obj.schema[defs_notation].keys():
        group = self.schema_obj.schema[defs_notation][group_key]
        for param_key, param in group["properties"].items():
            if param_key in ignore_params:
                ignored.append(f"Ignoring description check for param in schema: `{param_key}`")
                continue
            if "description" not in param.keys():
                warned.append(f"No description provided in schema for parameter: `{param_key}`")

    for ip in ignore_params:
        ignored.append(f"Parameter is ignored: `{ip}`")

    return {"passed": passed, "warned": warned, "ignored": ignored}



import os
import tempfile
from pathlib import Path
from unittest import mock

import pytest

import nf_core.components.components_command
import nf_core.components.patch
import nf_core.subworkflows

from ..test_subworkflows import TestSubworkflows
from ..utils import GITLAB_REPO, GITLAB_SUBWORKFLOWS_BRANCH, GITLAB_URL

OLD_SHA = "dbb12457e32d3da8eea7dc4ae096201fff4747c5"
SUCCEED_SHA = "0a33e6a0d730ad22a0ec9f7f9a7540af6e943221"
FAIL_SHA = "b6e5e8739de9a1a0c4f85267144e43dbaf8f1461"


class TestSubworkflowsPatch(TestSubworkflows):
    """
    Test the 'nf-core subworkflows patch' command
    """

    def modify_main_nf(self, path):
        """Modify a file to test patch creation"""
        with open(path) as fh:
            lines = fh.readlines()
        # We want a patch file that looks something like:
        # -    ch_fasta // channel: [ fasta ]
        for line_index in range(len(lines)):
            if lines[line_index] == "    ch_fasta // channel: [ fasta ]\n":
                to_pop = line_index
        lines.pop(to_pop)
        with open(path, "w") as fh:
            fh.writelines(lines)

    def setup_patch(self, pipeline_dir, modify_subworkflow):
        # Install the subworkflow bam_sort_stats_samtools
        install_obj = nf_core.subworkflows.SubworkflowInstall(
            pipeline_dir,
            prompt=False,
            force=False,
            remote_url=GITLAB_URL,
            branch=GITLAB_SUBWORKFLOWS_BRANCH,
            sha=OLD_SHA,
        )

        # Install the module
        install_obj.install("bam_sort_stats_samtools")

        if modify_subworkflow:
            # Modify the subworkflow
            subworkflow_path = Path(pipeline_dir, "subworkflows", GITLAB_REPO, "bam_sort_stats_samtools")
            self.modify_main_nf(subworkflow_path / "main.nf")

    def test_create_patch_no_change(self):
        """Test creating a patch when there is a change to the module"""
        self.setup_patch(self.pipeline_dir, False)

        # Try creating a patch file
        patch_obj = nf_core.subworkflows.SubworkflowPatch(self.pipeline_dir, GITLAB_URL, GITLAB_SUBWORKFLOWS_BRANCH)
        with pytest.raises(UserWarning):
            patch_obj.patch("bam_sort_stats_samtools")

        subworkflow_path = Path(self.pipeline_dir, "subworkflows", GITLAB_REPO, "bam_sort_stats_samtools")

        # Check that no patch file has been added to the directory
        assert not (subworkflow_path / "bam_sort_stats_samtools.diff").exists()

    def test_create_patch_change(self):
        """Test creating a patch when there is no change to the subworkflow"""
        self.setup_patch(self.pipeline_dir, True)

        # Try creating a patch file
        patch_obj = nf_core.subworkflows.SubworkflowPatch(self.pipeline_dir, GITLAB_URL, GITLAB_SUBWORKFLOWS_BRANCH)
        patch_obj.patch("bam_sort_stats_samtools")

        subworkflow_path = Path(self.pipeline_dir, "subworkflows", GITLAB_REPO, "bam_sort_stats_samtools")

        # Check that a patch file with the correct name has been created
        assert (subworkflow_path / "bam_sort_stats_samtools.diff").exists()

        # Check that the correct lines are in the patch file
        with open(subworkflow_path / "bam_sort_stats_samtools.diff") as fh:
            patch_lines = fh.readlines()
        print(patch_lines)
        subworkflow_relpath = subworkflow_path.relative_to(self.pipeline_dir)
        assert f"--- {subworkflow_relpath / 'main.nf'}\n" in patch_lines, subworkflow_relpath / "main.nf"
        assert f"+++ {subworkflow_relpath / 'main.nf'}\n" in patch_lines
        assert "-    ch_fasta // channel: [ fasta ]\n" in patch_lines

    def test_create_patch_try_apply_successful(self):
        """Test creating a patch file and applying it to a new version of the the files"""
        self.setup_patch(self.pipeline_dir, True)
        subworkflow_relpath = Path("subworkflows", GITLAB_REPO, "bam_sort_stats_samtools")
        subworkflow_path = Path(self.pipeline_dir, subworkflow_relpath)

        # Try creating a patch file
        patch_obj = nf_core.subworkflows.SubworkflowPatch(self.pipeline_dir, GITLAB_URL, GITLAB_SUBWORKFLOWS_BRANCH)
        patch_obj.patch("bam_sort_stats_samtools")

        # Check that a patch file with the correct name has been created
        assert (subworkflow_path / "bam_sort_stats_samtools.diff").exists()

        update_obj = nf_core.subworkflows.SubworkflowUpdate(
            self.pipeline_dir, sha=OLD_SHA, remote_url=GITLAB_URL, branch=GITLAB_SUBWORKFLOWS_BRANCH
        )

        # Install the new files
        install_dir = Path(tempfile.mkdtemp())
        update_obj.install_component_files("bam_sort_stats_samtools", OLD_SHA, update_obj.modules_repo, install_dir)

        # Try applying the patch
        subworkflow_install_dir = install_dir / "bam_sort_stats_samtools"
        patch_relpath = subworkflow_relpath / "bam_sort_stats_samtools.diff"
        assert (
            update_obj.try_apply_patch(
                "bam_sort_stats_samtools", GITLAB_REPO, patch_relpath, subworkflow_path, subworkflow_install_dir
            )
            is True
        )

        # Move the files from the temporary directory
        update_obj.move_files_from_tmp_dir("bam_sort_stats_samtools", install_dir, GITLAB_REPO, OLD_SHA)

        # Check that a patch file with the correct name has been created
        assert (subworkflow_path / "bam_sort_stats_samtools.diff").exists()

        # Check that the correct lines are in the patch file
        with open(subworkflow_path / "bam_sort_stats_samtools.diff") as fh:
            patch_lines = fh.readlines()
        subworkflow_relpath = subworkflow_path.relative_to(self.pipeline_dir)
        assert f"--- {subworkflow_relpath / 'main.nf'}\n" in patch_lines, subworkflow_relpath / "main.nf"
        assert f"+++ {subworkflow_relpath / 'main.nf'}\n" in patch_lines
        assert "-    ch_fasta // channel: [ fasta ]\n" in patch_lines

        # Check that 'main.nf' is updated correctly
        with open(subworkflow_path / "main.nf") as fh:
            main_nf_lines = fh.readlines()
        # These lines should have been removed by the patch
        assert "-    ch_fasta // channel: [ fasta ]\n" not in main_nf_lines

    def test_create_patch_try_apply_failed(self):
        """Test creating a patch file and applying it to a new version of the the files"""
        self.setup_patch(self.pipeline_dir, True)
        subworkflow_relpath = Path("subworkflows", GITLAB_REPO, "bam_sort_stats_samtools")
        subworkflow_path = Path(self.pipeline_dir, subworkflow_relpath)

        # Try creating a patch file
        patch_obj = nf_core.subworkflows.SubworkflowPatch(self.pipeline_dir, GITLAB_URL, GITLAB_SUBWORKFLOWS_BRANCH)
        patch_obj.patch("bam_sort_stats_samtools")

        # Check that a patch file with the correct name has been created
        assert (subworkflow_path / "bam_sort_stats_samtools.diff").exists()

        update_obj = nf_core.subworkflows.SubworkflowUpdate(
            self.pipeline_dir, remote_url=GITLAB_URL, branch=GITLAB_SUBWORKFLOWS_BRANCH
        )

        # Install the new files
        install_dir = Path(tempfile.mkdtemp())
        update_obj.install_component_files("bam_sort_stats_samtools", FAIL_SHA, update_obj.modules_repo, install_dir)

        # Try applying the patch
        subworkflow_install_dir = install_dir / "bam_sort_stats_samtools"
        patch_relpath = subworkflow_relpath / "bam_sort_stats_samtools.diff"
        assert (
            update_obj.try_apply_patch(
                "bam_sort_stats_samtools", GITLAB_REPO, patch_relpath, subworkflow_path, subworkflow_install_dir
            )
            is False
        )

    def test_create_patch_update_success(self):
        """
        Test creating a patch file and the updating the subworkflow

        Should have the same effect as 'test_create_patch_try_apply_successful'
        but uses higher level api
        """
        self.setup_patch(self.pipeline_dir, True)
        swf_path = Path(self.pipeline_dir, "subworkflows", GITLAB_REPO, "bam_sort_stats_samtools")

        # Try creating a patch file
        patch_obj = nf_core.subworkflows.SubworkflowPatch(self.pipeline_dir, GITLAB_URL, GITLAB_SUBWORKFLOWS_BRANCH)
        patch_obj.patch("bam_sort_stats_samtools")

        patch_fn = "bam_sort_stats_samtools.diff"
        # Check that a patch file with the correct name has been created
        assert (swf_path / patch_fn).exists()

        # Check the 'modules.json' contains a patch file for the subworkflow
        modules_json_obj = nf_core.modules.modules_json.ModulesJson(self.pipeline_dir)
        assert modules_json_obj.get_patch_fn(
            "subworkflows", "bam_sort_stats_samtools", GITLAB_URL, GITLAB_REPO
        ) == Path("subworkflows", GITLAB_REPO, "bam_sort_stats_samtools", patch_fn)

        # Update the subworkflow
        update_obj = nf_core.subworkflows.update.SubworkflowUpdate(
            self.pipeline_dir,
            sha=OLD_SHA,
            show_diff=False,
            update_deps=True,
            remote_url=GITLAB_URL,
            branch=GITLAB_SUBWORKFLOWS_BRANCH,
        )
        assert update_obj.update("bam_sort_stats_samtools")

        # Check that a patch file with the correct name has been created
        assert (swf_path / patch_fn).exists()

        # Check the 'modules.json' contains a patch file for the subworkflow
        modules_json_obj = nf_core.modules.modules_json.ModulesJson(self.pipeline_dir)
        assert modules_json_obj.get_patch_fn(
            "subworkflows", "bam_sort_stats_samtools", GITLAB_URL, GITLAB_REPO
        ) == Path("subworkflows", GITLAB_REPO, "bam_sort_stats_samtools", patch_fn), modules_json_obj.get_patch_fn(
            "subworkflows", "bam_sort_stats_samtools", GITLAB_URL, GITLAB_REPO
        )

        # Check that the correct lines are in the patch file
        with open(swf_path / patch_fn) as fh:
            patch_lines = fh.readlines()
        swf_relpath = swf_path.relative_to(self.pipeline_dir)
        assert f"--- {swf_relpath / 'main.nf'}\n" in patch_lines
        assert f"+++ {swf_relpath / 'main.nf'}\n" in patch_lines
        assert "-    ch_fasta // channel: [ fasta ]\n" in patch_lines

        # Check that 'main.nf' is updated correctly
        with open(swf_path / "main.nf") as fh:
            main_nf_lines = fh.readlines()
        # this line should have been removed by the patch
        assert "    ch_fasta // channel: [ fasta ]\n" not in main_nf_lines

    def test_create_patch_update_fail(self):
        """
        Test creating a patch file and updating a subworkflow when there is a diff conflict
        """
        self.setup_patch(self.pipeline_dir, True)
        swf_path = Path(self.pipeline_dir, "subworkflows", GITLAB_REPO, "bam_sort_stats_samtools")

        # Try creating a patch file
        patch_obj = nf_core.subworkflows.SubworkflowPatch(self.pipeline_dir, GITLAB_URL, GITLAB_SUBWORKFLOWS_BRANCH)
        patch_obj.patch("bam_sort_stats_samtools")

        patch_fn = "bam_sort_stats_samtools.diff"
        # Check that a patch file with the correct name has been created
        assert (swf_path / patch_fn).exists()

        # Check the 'modules.json' contains a patch file for the subworkflow
        modules_json_obj = nf_core.modules.modules_json.ModulesJson(self.pipeline_dir)
        assert modules_json_obj.get_patch_fn(
            "subworkflows", "bam_sort_stats_samtools", GITLAB_URL, GITLAB_REPO
        ) == Path("subworkflows", GITLAB_REPO, "bam_sort_stats_samtools", patch_fn)

        # Save the file contents for downstream comparison
        with open(swf_path / patch_fn) as fh:
            patch_contents = fh.read()

        update_obj = nf_core.subworkflows.update.SubworkflowUpdate(
            self.pipeline_dir,
            sha=FAIL_SHA,
            show_diff=False,
            update_deps=True,
            remote_url=GITLAB_URL,
            branch=GITLAB_SUBWORKFLOWS_BRANCH,
        )
        update_obj.update("bam_sort_stats_samtools")

        # Check that the installed files have not been affected by the attempted patch
        temp_dir = Path(tempfile.mkdtemp())
        nf_core.components.components_command.ComponentCommand(
            "subworkflows", self.pipeline_dir, GITLAB_URL, GITLAB_SUBWORKFLOWS_BRANCH
        ).install_component_files("bam_sort_stats_samtools", FAIL_SHA, update_obj.modules_repo, temp_dir)

        temp_module_dir = temp_dir / "bam_sort_stats_samtools"
        for file in os.listdir(temp_module_dir):
            assert file in os.listdir(swf_path)
            with open(swf_path / file) as fh:
                installed = fh.read()
            with open(temp_module_dir / file) as fh:
                shouldbe = fh.read()
            assert installed == shouldbe

        # Check that the patch file is unaffected
        with open(swf_path / patch_fn) as fh:
            new_patch_contents = fh.read()
        assert patch_contents == new_patch_contents

    def test_remove_patch(self):
        """Test creating a patch when there is no change to the subworkflow"""
        self.setup_patch(self.pipeline_dir, True)

        # Try creating a patch file
        patch_obj = nf_core.subworkflows.SubworkflowPatch(self.pipeline_dir, GITLAB_URL, GITLAB_SUBWORKFLOWS_BRANCH)
        patch_obj.patch("bam_sort_stats_samtools")

        subworkflow_path = Path(self.pipeline_dir, "subworkflows", GITLAB_REPO, "bam_sort_stats_samtools")

        # Check that a patch file with the correct name has been created
        assert (subworkflow_path / "bam_sort_stats_samtools.diff").exists()

        with mock.patch.object(nf_core.components.patch.questionary, "confirm") as mock_questionary:
            mock_questionary.unsafe_ask.return_value = True
            patch_obj.remove("bam_sort_stats_samtools")
        # Check that the diff file has been removed
        assert not (subworkflow_path / "bam_sort_stats_samtools.diff").exists()



from .create import SubworkflowCreate
from .info import SubworkflowInfo
from .install import SubworkflowInstall
from .lint import SubworkflowLint
from .list import SubworkflowList
from .patch import SubworkflowPatch
from .remove import SubworkflowRemove
from .update import SubworkflowUpdate



import json
import logging
from pathlib import Path
from typing import Dict, List, Optional, Union, cast

import rich.table

from nf_core.components.components_command import ComponentCommand
from nf_core.modules.modules_json import ModulesJson, ModulesJsonModuleEntry
from nf_core.modules.modules_repo import ModulesRepo

log = logging.getLogger(__name__)


class ComponentList(ComponentCommand):
    def __init__(
        self,
        component_type: str,
        pipeline_dir: Union[str, Path] = ".",
        remote: bool = True,
        remote_url: Optional[str] = None,
        branch: Optional[str] = None,
        no_pull: bool = False,
    ) -> None:
        self.remote = remote
        super().__init__(component_type, pipeline_dir, remote_url, branch, no_pull)

    def _configure_repo_and_paths(self, nf_dir_req: bool = True) -> None:
        """
        Override the default with nf_dir_req set to False to allow
        info to be run from anywhere and still return remote info
        """
        if self.remote:
            nf_dir_req = False
        return super()._configure_repo_and_paths(nf_dir_req)

    def list_components(
        self, keywords: Optional[List[str]] = None, print_json: bool = False
    ) -> Union[rich.table.Table, str]:
        keywords = keywords or []
        """
        Get available modules/subworkflows names from GitHub tree for repo
        and print as list to stdout
        """
        # Check modules directory structure
        # self.check_component_structure(self.component_type)

        # Initialise rich table
        table: rich.table.Table = rich.table.Table()
        table.add_column(f"{self.component_type[:-1].capitalize()} Name")
        components: List[str] = []

        def pattern_msg(keywords: List[str]) -> str:
            if len(keywords) == 0:
                return ""
            if len(keywords) == 1:
                return f" matching pattern '{keywords[0]}'"
            else:
                quoted_keywords = (f"'{key}'" for key in keywords)
                return f" matching patterns {', '.join(quoted_keywords)}"

        # No pipeline given - show all remote
        if self.remote:
            # Filter the modules/subworkflows by keywords
            components = [
                comp
                for comp in self.modules_repo.get_avail_components(self.component_type)
                if all(k in comp for k in keywords)
            ]

            # Nothing found
            if len(components) == 0:
                log.info(
                    f"No available {self.component_type} found in {self.modules_repo.remote_url} ({self.modules_repo.branch})"
                    f"{pattern_msg(keywords)}"
                )
                return ""

            for comp in sorted(components):
                table.add_row(comp)

        # We have a pipeline - list what's installed
        else:
            # Check that we are in a pipeline directory
            log.info(f"Repository type: [blue]{self.repo_type}")
            try:
                if self.repo_type != "pipeline":
                    raise UserWarning(
                        f"The command 'nf-core {self.component_type} list local' must be run from a pipeline directory.",
                    )
            except UserWarning as e:
                log.error(e)
                return ""
            # Check whether pipelines is valid
            try:
                self.has_valid_directory()
            except UserWarning as e:
                log.error(e)
                return ""

            # Verify that 'modules.json' is consistent with the installed modules
            modules_json: ModulesJson = ModulesJson(self.directory)
            modules_json.check_up_to_date()

            # Filter by keywords
            repos_with_comps = {
                repo_url: [comp for comp in components if all(k in comp[1] for k in keywords)]
                for repo_url, components in modules_json.get_all_components(self.component_type).items()
            }

            # Nothing found
            if sum(map(len, repos_with_comps)) == 0:
                log.info(f"No nf-core {self.component_type} found in '{self.directory}'{pattern_msg(keywords)}")
                return ""

            table.add_column("Repository")
            table.add_column("Version SHA")
            table.add_column("Message")
            table.add_column("Date")

            # Load 'modules.json'
            modules_json_file = modules_json.modules_json

            for repo_url, component_with_dir in sorted(repos_with_comps.items()):
                repo_entry: Dict[str, Dict[str, Dict[str, ModulesJsonModuleEntry]]]
                if modules_json_file is None:
                    log.warning(f"Modules JSON file '{modules_json.modules_json_path}' is missing. ")
                    continue
                else:
                    repo_entry = modules_json_file["repos"].get(repo_url, {})
                    for install_dir, component in sorted(component_with_dir):
                        # Use cast() to predict the return type of recursive get():s
                        repo_modules = cast(dict, repo_entry.get(self.component_type))
                        component_entry = cast(dict, cast(dict, repo_modules.get(install_dir)).get(component))

                        if component_entry:
                            version_sha = component_entry["git_sha"]
                            try:
                                # pass repo_name to get info on modules even outside nf-core/modules
                                module = ModulesRepo(
                                    remote_url=repo_url,
                                    branch=component_entry["branch"],
                                )
                                message, date = module.get_commit_info(version_sha)
                            except LookupError as e:
                                log.warning(e)
                                date = "[red]Not Available"
                                message = "[red]Not Available"
                        else:
                            log.warning(
                                f"Commit SHA for {self.component_type[:-1]} '{install_dir}/{self.component_type}' is missing from 'modules.json'"
                            )
                            version_sha = "[red]Not Available"
                            date = "[red]Not Available"
                            message = "[red]Not Available"
                        nice_repo_name = repo_url.replace("https://github.com/", "").replace(".git", "")
                        table.add_row(
                            component,
                            f"[link={repo_url}]{nice_repo_name}[/link]",
                            f"[link={module.gitless_repo()}/commit/{version_sha}]{version_sha[:7]}[/link]",
                            message,
                            date,
                        )
                        components.append(component)

        if print_json:
            return json.dumps(components, sort_keys=True, indent=4)

        if self.remote:
            log.info(
                f"{self.component_type.capitalize()} available from {self.modules_repo.remote_url} ({self.modules_repo.branch})"
                f"{pattern_msg(keywords)}:\n"
            )
        else:
            log.info(f"{self.component_type.capitalize()} installed in '{self.directory}'{pattern_msg(keywords)}:\n")
        return table



import json
from pathlib import Path

import yaml
from rich.console import Console

import nf_core.modules.list

from ..test_modules import TestModules
from ..utils import GITLAB_DEFAULT_BRANCH, GITLAB_URL


class TestModulesList(TestModules):
    def test_modules_list_remote(self):
        """Test listing available modules"""
        mods_list = nf_core.modules.list.ModuleList()
        listed_mods = mods_list.list_components()
        console = Console(record=True)
        console.print(listed_mods)
        output = console.export_text()
        assert "fastqc" in output

    def test_modules_list_remote_gitlab(self):
        """Test listing the modules in the remote gitlab repo"""
        mods_list = nf_core.modules.list.ModuleList(remote_url=GITLAB_URL, branch=GITLAB_DEFAULT_BRANCH)
        listed_mods = mods_list.list_components()
        console = Console(record=True)
        console.print(listed_mods)
        output = console.export_text()
        assert "fastqc" in output

    def test_modules_list_pipeline(self):
        """Test listing locally installed modules"""
        mods_list = nf_core.modules.list.ModuleList(self.pipeline_dir, remote=False)
        listed_mods = mods_list.list_components()
        console = Console(record=True)
        console.print(listed_mods)
        output = console.export_text()
        assert "fastqc" in output
        assert "multiqc" in output

    def test_modules_install_and_list_pipeline(self):
        """Test listing locally installed modules"""
        self.mods_install.install("trimgalore")
        mods_list = nf_core.modules.list.ModuleList(self.pipeline_dir, remote=False)
        listed_mods = mods_list.list_components()
        console = Console(record=True)
        console.print(listed_mods)
        output = console.export_text()
        assert "trimgalore" in output

    def test_modules_install_gitlab_and_list_pipeline(self):
        """Test listing locally installed modules"""
        self.mods_install_gitlab.install("fastqc")
        mods_list = nf_core.modules.list.ModuleList(self.pipeline_dir, remote=False)
        listed_mods = mods_list.list_components()
        console = Console(record=True)
        console.print(listed_mods)
        output = console.export_text()
        assert "fastqc" in output

    def test_modules_list_local_json(self):
        """Test listing locally installed modules as JSON"""
        mods_list = nf_core.modules.list.ModuleList(self.pipeline_dir, remote=False)
        listed_mods = str(mods_list.list_components(print_json=True))
        listed_mods = json.loads(listed_mods)
        assert "fastqc" in listed_mods
        assert "multiqc" in listed_mods

    def test_modules_list_remote_json(self) -> None:
        """Test listing available modules as JSON"""
        mods_list = nf_core.modules.list.ModuleList(remote=True)
        listed_mods: str = str(mods_list.list_components(print_json=True))
        listed_mods = json.loads(listed_mods)
        assert "fastqc" in listed_mods
        assert "multiqc" in listed_mods

    def test_modules_list_with_one_keyword(self):
        """Test listing available modules with one keyword"""
        mods_list = nf_core.modules.list.ModuleList(remote=True)
        listed_mods = mods_list.list_components(keywords=["qc"])
        console = Console(record=True)
        console.print(listed_mods)
        output = console.export_text()
        assert "multiqc" in output

    def test_modules_list_with_keywords(self):
        """Test listing available modules with multiple keywords"""
        mods_list = nf_core.modules.list.ModuleList(remote=True)
        listed_mods = mods_list.list_components(keywords=["fastq", "qc"])
        console = Console(record=True)
        console.print(listed_mods)
        output = console.export_text()
        assert "fastqc" in output

    def test_modules_list_with_unused_keyword(self):
        """Test listing available modules with an unused keyword"""
        mods_list = nf_core.modules.list.ModuleList(remote=True)
        with self.assertLogs(level="INFO") as log:
            listed_mods = mods_list.list_components(keywords=["you_will_never_find_me"])
            self.assertIn("No available", log.output[0])
            # expect empty list
            assert listed_mods == ""

    def test_modules_list_in_wrong_repo_fail(self):
        """Test listing available modules in a non-pipeline repo"""
        # modify repotype in .nf-core.yml
        with open(Path(self.pipeline_dir, ".nf-core.yml")) as fh:
            nf_core_yml = yaml.safe_load(fh)
        nf_core_yml_orig = nf_core_yml.copy()
        nf_core_yml["repository_type"] = "modules"
        nf_core_yml["org_path"] = "nf-core"

        print(nf_core_yml)
        with open(Path(self.pipeline_dir, ".nf-core.yml"), "w") as fh:
            yaml.safe_dump(nf_core_yml, fh)
        # expect error logged
        with self.assertLogs(level="ERROR") as log:
            mods_list = nf_core.modules.list.ModuleList(self.pipeline_dir, remote=False)
            listed_mods = mods_list.list_components()
            self.assertIn("must be run from a pipeline directory", log.output[0])
            # expect empty list
            assert listed_mods == ""
        # restore .nf-core.yml
        with open(Path(self.pipeline_dir, ".nf-core.yml"), "w") as fh:
            yaml.safe_dump(nf_core_yml_orig, fh)



import logging

from nf_core.components.create import ComponentCreate

log = logging.getLogger(__name__)


class SubworkflowCreate(ComponentCreate):
    def __init__(
        self,
        pipeline_dir,
        component="",
        author=None,
        force=False,
        migrate_pytest=False,
    ):
        super().__init__(
            "subworkflows",
            pipeline_dir,
            component,
            author,
            force=force,
            migrate_pytest=migrate_pytest,
        )



import logging
from pathlib import Path

import questionary
from rich.console import Console, Group
from rich.panel import Panel
from rich.syntax import Syntax

import nf_core.utils
from nf_core.components.components_command import ComponentCommand
from nf_core.modules.modules_json import ModulesJson

from .install import ComponentInstall

log = logging.getLogger(__name__)


class ComponentRemove(ComponentCommand):
    def __init__(self, component_type, pipeline_dir, remote_url=None, branch=None, no_pull=False):
        super().__init__(component_type, pipeline_dir, remote_url, branch, no_pull)

    def remove(self, component, removed_by=None, removed_components=None, force=False):
        """
        Remove an already installed module/subworkflow
        This command only works for modules/subworkflows that are installed from 'nf-core/modules'

        Args:
            component (str): Name of the component to remove
            removed_by (str): Name of the component that is removing the current component
                (a subworkflow name if the component is a dependency or "modules" or "subworkflows" if it is not a dependency)
            removed_components (list[str]): list of components that have been removed during a recursive remove of subworkflows
            force (bool): Force removal of component, even if there is still an include statement in a workflow file

        Returns:
            bool: True if any item has been removed, False if not
        """
        if self.repo_type == "modules":
            log.error(f"You cannot remove a {self.component_type[:-1]} in a clone of nf-core/modules")
            return False

        # Check modules directory structure
        if self.component_type == "modules":
            self.check_modules_structure()

        # Check whether pipeline is valid and with a modules.json file
        self.has_valid_directory()
        self.has_modules_file()

        repo_path = self.modules_repo.repo_path
        if component is None:
            component = questionary.autocomplete(
                f"{self.component_type[:-1]} name:",
                choices=self.components_from_repo(repo_path),
                style=nf_core.utils.nfcore_question_style,
            ).unsafe_ask()

        if removed_components is None:
            removed_components = []

        # Get the module/subworkflow directory
        component_dir = Path(self.directory, self.component_type, repo_path, component)

        # Load the modules.json file
        modules_json = ModulesJson(self.directory)
        modules_json.load()

        # Verify that the module/subworkflow is actually installed
        if not component_dir.exists():
            log.error(f"Installation directory '{component_dir}' does not exist.")

            if modules_json.component_present(component, self.modules_repo.remote_url, repo_path, self.component_type):
                log.error(f"Found entry for '{component}' in 'modules.json'. Removing...")
                modules_json.remove_entry(self.component_type, component, self.modules_repo.remote_url, repo_path)
            return False

        # remove all dependent components based on installed_by entry
        # Remove entry from modules.json
        removed = False
        removed_components = []
        # Remove component from modules.json
        removed_component = modules_json.remove_entry(
            self.component_type,
            component,
            self.modules_repo.remote_url,
            repo_path,
            removed_by=removed_by,
        )
        removed_component_dir = Path(self.component_type, repo_path, component)
        if removed_component:
            # check if the module/subworkflow has been manually included in the pipeline
            include_stmts = self.check_if_in_include_stmts(str(removed_component_dir))
            if include_stmts:
                # print the include statements
                log.warning(
                    f"The {self.component_type[:-1]} '{component}' is still included in the following workflow file{nf_core.utils.plural_s(include_stmts)}:"
                )
                console = Console()
                for file, stmts in include_stmts.items():
                    renderables = []
                    for stmt in stmts:
                        # check that the line number is integer
                        if not isinstance(stmt["line_number"], int):
                            log.error(
                                f"Could not parse line number '{stmt['line_number']}' in '{file}'. Please report this issue."
                            )
                            continue

                        renderables.append(
                            Syntax(
                                str(stmt["line"]),
                                "groovy",
                                theme="ansi_dark",
                                line_numbers=True,
                                start_line=stmt["line_number"],
                            )
                        )
                    console.print(
                        Panel(
                            Group(*renderables),
                            title=f"{file}",
                            style="white",
                            title_align="center",
                            padding=1,
                        )
                    )
                # ask the user if they still want to remove the component, add it back otherwise
                if not force:
                    if not questionary.confirm(
                        f"Do you still want to remove the {self.component_type[:-1]} '{component}'?",
                        style=nf_core.utils.nfcore_question_style,
                    ).unsafe_ask():
                        # add the component back to modules.json
                        if not ComponentInstall(self.directory, self.component_type, force=True).install(
                            component, silent=True
                        ):
                            log.warning(
                                f"Could not install the {self.component_type[:-1]} '{component}', please install it manually with 'nf-core {self.component_type} install  {component}'."
                            )
                        removed_components.append(component)
                        return removed
            # Remove the component files of all entries removed from modules.json
            removed = (
                True
                if self.clear_component_dir(component, Path(self.directory, removed_component_dir)) or removed
                else False
            )
            removed_components.append(component)

        if removed:
            if self.component_type == "subworkflows":
                removed_by = component
                dependent_components = modules_json.get_dependent_components(
                    self.component_type, component, self.modules_repo.remote_url, repo_path, {}
                )
                for component_name, component_type in dependent_components.items():
                    if component_name in removed_components:
                        continue
                    original_component_type = self.component_type
                    self.component_type = component_type
                    dependency_removed = self.remove(
                        component_name, removed_by=removed_by, removed_components=removed_components
                    )
                    self.component_type = original_component_type
                    # remember removed dependencies
                    if dependency_removed:
                        removed_components.append(component_name.replace("/", "_"))
            # print removed dependencies
            if removed_components:
                log.info(f"Removed files for '{component}' and its dependencies '{', '.join(removed_components)}'.")
            else:
                log.info(f"Removed files for '{component}'.")
        else:
            installed_by = modules_json.get_installed_by_entries(self.component_type, component)
            if installed_by == [self.component_type]:
                log.error(
                    f"Did not remove '{component}', because it was also manually installed. Only updated 'installed_by' entry in modules.json."
                )
            log.info(
                f"""Did not remove {self.component_type[:-1]} '{component}', because it was also installed by {', '.join(f"'{d}'" for d in installed_by)}. Only updated the 'installed_by' entry in modules.json."""
            )
        return removed



from pathlib import Path

import nf_core.pipelines.lint

from ..test_lint import TestLint


class TestLintFilesUnchanged(TestLint):
    def test_files_unchanged_pass(self):
        self.lint_obj._load()
        results = self.lint_obj.files_unchanged()
        assert len(results.get("warned", [])) == 0
        assert len(results.get("failed", [])) == 0
        assert len(results.get("ignored", [])) == 0
        assert not results.get("could_fix", True)

    def test_files_unchanged_fail(self):
        failing_file = Path(".github", "CONTRIBUTING.md")
        new_pipeline = self._make_pipeline_copy()
        with open(Path(new_pipeline, failing_file), "a") as fh:
            fh.write("THIS SHOULD NOT BE HERE")

        lint_obj = nf_core.pipelines.lint.PipelineLint(new_pipeline)
        lint_obj._load()
        results = lint_obj.files_unchanged()
        assert len(results["failed"]) > 0
        assert str(failing_file) in results["failed"][0]
        assert results["could_fix"]



"""Some tests covering the linting code."""

import json
from pathlib import Path

import yaml

import nf_core.pipelines.create.create
import nf_core.pipelines.lint

from ..test_pipelines import TestPipelines
from ..utils import with_temporary_folder


class TestLint(TestPipelines):
    """Class for lint tests"""

    def setUp(self) -> None:
        super().setUp()
        self.lint_obj = nf_core.pipelines.lint.PipelineLint(self.pipeline_dir)


##########################
# CORE lint.py FUNCTIONS #
##########################
class TestPipelinesLint(TestLint):
    def test_run_linting_function(self):
        """Run the run_linting() function in lint.py

        We don't really check any of this code as it's just a series of function calls
        and we're testing each of those individually. This is mostly to check for syntax errors."""
        nf_core.pipelines.lint.run_linting(self.pipeline_dir, False)

    def test_init_pipeline_lint(self):
        """Simply create a PipelineLint object.

        This checks that all of the lint test imports are working properly,
        we also check that the git sha was found and that the release flag works properly
        """
        lint_obj = nf_core.pipelines.lint.PipelineLint(self.pipeline_dir, True)

        # Tests that extra test is added for release mode
        assert "version_consistency" in lint_obj.lint_tests
        assert lint_obj.git_sha
        # Tests that parent nf_core.utils.Pipeline class __init__() is working to find git hash
        assert len(lint_obj.git_sha) > 0

    def test_load_lint_config_not_found(self):
        """Try to load a linting config file that doesn't exist"""
        assert self.lint_obj._load_lint_config()
        assert self.lint_obj.lint_config is not None
        assert self.lint_obj.lint_config.model_dump(exclude_none=True) == {}

    def test_load_lint_config_ignore_all_tests(self):
        """Try to load a linting config file that ignores all tests"""

        # Make a copy of the test pipeline and create a lint object
        new_pipeline = self._make_pipeline_copy()
        lint_obj = nf_core.pipelines.lint.PipelineLint(new_pipeline)

        # Make a config file listing all test names
        config_dict = {"repository_type": "pipeline", "lint": {test_name: False for test_name in lint_obj.lint_tests}}
        with open(Path(new_pipeline, ".nf-core.yml"), "w") as fh:
            yaml.dump(config_dict, fh)

        # Load the new lint config file and check
        lint_obj._load_lint_config()
        assert lint_obj.lint_config is not None
        assert sorted(list(lint_obj.lint_config.model_dump(exclude_none=True))) == sorted(lint_obj.lint_tests)

        # Try running linting and make sure that all tests are ignored
        lint_obj._lint_pipeline()
        assert len(lint_obj.passed) == 0
        assert len(lint_obj.warned) == 0
        assert len(lint_obj.failed) == 0
        assert len(lint_obj.ignored) == len(lint_obj.lint_tests)

    @with_temporary_folder
    def test_json_output(self, tmp_dir):
        """
        Test creation of a JSON file with lint results

        Expected JSON output:
        {
            "nf_core_tools_version": "1.10.dev0",
            "date_run": "2020-06-05 10:56:42",
            "tests_pass": [
                [ 1, "This test passed"],
                [ 2, "This test also passed"]
            ],
            "tests_warned": [
                [ 2, "This test gave a warning"]
            ],
            "tests_failed": [],
            "num_tests_pass": 2,
            "num_tests_warned": 1,
            "num_tests_failed": 0,
            "has_tests_pass": true,
            "has_tests_warned": true,
            "has_tests_failed": false
        }
        """
        self.lint_obj.passed.append(("test_one", "This test passed"))
        self.lint_obj.passed.append(("test_two", "This test also passed"))
        self.lint_obj.warned.append(("test_three", "This test gave a warning"))

        # Make a temp dir for the JSON output
        json_fn = Path(tmp_dir, "lint_results.json")
        self.lint_obj._save_json_results(json_fn)

        # Load created JSON file and check its contents
        with open(json_fn) as fh:
            try:
                saved_json = json.load(fh)
            except json.JSONDecodeError as e:
                raise UserWarning(f"Unable to load JSON file '{json_fn}' due to error {e}")
        assert saved_json["num_tests_pass"] > 0
        assert saved_json["num_tests_warned"] > 0
        assert saved_json["num_tests_ignored"] == 0
        assert saved_json["num_tests_failed"] == 0
        assert saved_json["has_tests_pass"]
        assert saved_json["has_tests_warned"]
        assert not saved_json["has_tests_ignored"]
        assert not saved_json["has_tests_failed"]

    def test_wrap_quotes(self):
        md = self.lint_obj._wrap_quotes(["one", "two", "three"])
        assert md == "`one` or `two` or `three`"

    def test_sphinx_md_files(self):
        """Check that we have .md files for all lint module code,
        and that there are no unexpected files (eg. deleted lint tests)"""

        docs_basedir = Path(Path(__file__).parent.parent.parent, "docs", "api", "_src", "pipeline_lint_tests")

        # Get list of existing .md files
        existing_docs = []
        existing_docs = [
            str(Path(docs_basedir, fn))
            for fn in Path(docs_basedir).iterdir()
            if fn.match("*.md") and not fn.match("index.md")
        ]

        # Check .md files against each test name
        lint_obj = nf_core.pipelines.lint.PipelineLint("", True)
        for test_name in lint_obj.lint_tests:
            fn = Path(docs_basedir, f"{test_name}.md")
            assert fn.exists(), f"Could not find lint docs .md file: {fn}"
            existing_docs.remove(str(fn))

        # Check that we have no remaining .md files that we didn't expect
        assert len(existing_docs) == 0, f"Unexpected lint docs .md files found: {', '.join(existing_docs)}"



import logging

from nf_core.components.patch import ComponentPatch

log = logging.getLogger(__name__)


class SubworkflowPatch(ComponentPatch):
    def __init__(self, pipeline_dir, remote_url=None, branch=None, no_pull=False, installed_by=False):
        super().__init__(pipeline_dir, "subworkflows", remote_url, branch, no_pull, installed_by)



"""
Update a nextflow.config file with refgenie genomes
"""

import logging
import os
import re
from pathlib import Path
from textwrap import dedent

import rich
import rich.traceback
import yaml

import nf_core.utils

# import refgenconf


# Set up logging
log = logging.getLogger(__name__)
log.setLevel(logging.INFO)

# # Setup rich traceback
stderr = rich.console.Console(stderr=True, force_terminal=nf_core.utils.rich_force_colors())
rich.traceback.install(console=stderr, width=200, word_wrap=True, extra_lines=1)

NF_CFG_TEMPLATE = """
// This is a read-only config file managed by refgenie. Manual changes to this file will be overwritten
// To make changes here, use refgenie to update the reference genome data
params {{
  genomes {{
{content}
  }}
}}
"""


def _print_nf_config(rgc):
    """
    Generate a nextflow config file with the genomes
    from the refgenie config file
    Adapted from: https://github.com/refgenie/refgenie_nfcore

    Takes a RefGenConf object as argument
    """
    abg = rgc.list_assets_by_genome()
    genomes_str = ""
    alias_translations = _get_alias_translation_file(rgc)
    for genome, asset_list in abg.items():
        genomes_str += f"    '{genome}' {{\n"
        for asset in asset_list:
            try:
                pth = rgc.seek(genome, asset)
            # Catch general exception instead of refgencof exception --> no refgenconf import needed
            except Exception:
                log.warning(f"{genome}/{asset} is incomplete, ignoring...")
            else:
                # Translate an alias name to the alias used in the pipeline
                if asset in alias_translations.keys():
                    log.info(f"Translating refgenie asset alias {asset} to {alias_translations[asset]}.")
                    asset = alias_translations[asset]
                genomes_str += f'      {asset.ljust(20, " ")} = "{pth}"\n'
        genomes_str += "    }\n"

    return NF_CFG_TEMPLATE.format(content=genomes_str)


def _update_nextflow_home_config(refgenie_genomes_config_file, nxf_home):
    """
    Update the $NXF_HOME/config file by adding a includeConfig statement to it
    for the 'refgenie_genomes_config_file' if not already defined
    """
    # Check if NXF_HOME/config exists and has a
    include_config_string = dedent(
        f"""
        ///// >>> nf-core + RefGenie >>> /////
        // !! Contents within this block are managed by 'nf-core/tools' !!
        // Includes auto-generated config file with RefGenie genome assets
        includeConfig '{os.path.abspath(refgenie_genomes_config_file)}'
        ///// <<< nf-core + RefGenie <<< /////
        """
    )
    nxf_home_config = Path(nxf_home) / "config"
    if os.path.exists(nxf_home_config):
        # look for include statement in config
        has_include_statement = False
        with open(nxf_home_config) as fh:
            lines = fh.readlines()
            for line in lines:
                if re.match(rf"\s*includeConfig\s*'{os.path.abspath(refgenie_genomes_config_file)}'", line):
                    has_include_statement = True
                    break

        # if include statement is missing, add it to the last line
        if not has_include_statement:
            with open(nxf_home_config, "a") as fh:
                fh.write(include_config_string)

            log.info(f"Included refgenie_genomes.config to {nxf_home_config}")

    else:
        # create new config and add include statement
        with open(nxf_home_config, "w") as fh:
            fh.write(include_config_string)
            log.info(f"Created new nextflow config file: {nxf_home_config}")


def _get_alias_translation_file(rgc):
    """
    Read a file containing alias translations.

    Alias translation file should be located in the same folder as the refgenie `genome_config.yaml` file,
    the path is set to $REFGENIE environment variable by `refgenie init`.
    Alias translation file should be named `alias_translations.yaml`

    Input file contains the name of refgenie server aliases as keys and the name of the respective nf-core pipeline aliases as values.
    Such as:
    ensembl_gtf: gtf
    star_index: star
    """
    translations = {}

    if "REFGENIE" in os.environ:
        refgenie_genomes_config_path = os.environ.get("REFGENIE")
        refgenie_genomes_config_directory = Path(refgenie_genomes_config_path).parents[0]
    elif "genome_folder" in rgc:
        refgenie_genomes_config_directory = Path(rgc["genome_folder"])
    else:
        return translations

    try:
        with open(refgenie_genomes_config_directory / "alias_translations.yaml") as yaml_file:
            translations = yaml.load(yaml_file, Loader=yaml.Loader)
    except FileNotFoundError:
        pass

    return translations


def update_config(rgc):
    """
    Update the genomes.config file after a local refgenie database has been updated

    This function is executed after running 'refgenie pull <genome>/<asset>'
    The refgenie config file is transformed into a nextflow.config file, which is used to
    overwrite the 'refgenie_genomes.config' file.
    The path to the target  config file is inferred from the following options, in order:

    - the 'nextflow_config' attribute in the refgenie config file
    - the NXF_REFGENIE_PATH environment variable
    - otherwise defaults to: $NXF_HOME/nf-core/refgenie_genomes.config

    Additionally, an 'includeConfig' statement is added to the file $NXF_HOME/config
    """

    # Compile nextflow refgenie_genomes.config from refgenie config
    refgenie_genomes = _print_nf_config(rgc)

    # Get the path to NXF_HOME
    # If NXF_HOME is not set, create it at $HOME/.nextflow
    # If $HOME is not set, set nxf_home to false
    nxf_home = os.environ.get("NXF_HOME")
    if not nxf_home:
        try:
            nxf_home = Path.home() / ".nextflow"
            if not os.path.exists(nxf_home):
                log.info(f"Creating NXF_HOME directory at {nxf_home}")
                os.makedirs(nxf_home, exist_ok=True)
        except RuntimeError:
            nxf_home = False

    # Get the path for storing the updated refgenie_genomes.config
    if hasattr(rgc, "nextflow_config"):
        refgenie_genomes_config_file = rgc.nextflow_config
    elif "NXF_REFGENIE_PATH" in os.environ:
        refgenie_genomes_config_file = os.environ.get("NXF_REFGENIE_PATH")
    elif nxf_home:
        refgenie_genomes_config_file = Path(nxf_home) / "nf-core/refgenie_genomes.config"
    else:
        log.info("Could not determine path to 'refgenie_genomes.config' file.")
        return False

    if refgenie_genomes_config_file is None:
        log.info("Could not determine path to 'refgenie_genomes.config' file.")
        return False

    # Save the updated genome config
    try:
        with open(refgenie_genomes_config_file, "w+") as fh:
            fh.write(refgenie_genomes)
        log.info(f"Updated nf-core genomes config: {refgenie_genomes_config_file}")
    except FileNotFoundError:
        log.info(f"Could not write to {refgenie_genomes_config_file}")
        return False

    # Add include statement to NXF_HOME/config
    if nxf_home:
        _update_nextflow_home_config(refgenie_genomes_config_file, nxf_home)

    return True



"""
Helper functions for tests
"""

import filecmp
import functools
import tempfile
from pathlib import Path
from typing import Any, Callable, Tuple

import responses
import yaml

import nf_core.modules
import nf_core.pipelines.create.create
from nf_core import __version__
from nf_core.utils import NFCoreTemplateConfig, NFCoreYamlConfig, custom_yaml_dumper

TEST_DATA_DIR = Path(__file__).parent / "data"
OLD_TRIMGALORE_SHA = "9b7a3bdefeaad5d42324aa7dd50f87bea1b04386"
OLD_TRIMGALORE_BRANCH = "mimic-old-trimgalore"
GITLAB_URL = "https://gitlab.com/nf-core/modules-test.git"
GITLAB_REPO = "nf-core-test"
GITLAB_DEFAULT_BRANCH = "main"
GITLAB_SUBWORKFLOWS_BRANCH = "subworkflows"
GITLAB_SUBWORKFLOWS_ORG_PATH_BRANCH = "subworkflows-org-path"
OLD_SUBWORKFLOWS_SHA = "f3c078809a2513f1c95de14f6633fe1f03572fdb"
# Branch test stuff
GITLAB_BRANCH_TEST_BRANCH = "branch-tester"
GITLAB_BRANCH_ORG_PATH_BRANCH = "org-path"
GITLAB_BRANCH_TEST_OLD_SHA = "e772abc22c1ff26afdf377845c323172fb3c19ca"
GITLAB_BRANCH_TEST_NEW_SHA = "7d73e21f30041297ea44367f2b4fd4e045c0b991"
GITLAB_NFTEST_BRANCH = "nf-test-tests-self-hosted-runners"


def with_temporary_folder(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    Call the decorated function under the tempfile.TemporaryDirectory
    context manager. Pass the temporary directory name to the decorated
    function
    """

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        with tempfile.TemporaryDirectory() as tmpdirname:
            return func(*args, tmpdirname, **kwargs)

    return wrapper


def with_temporary_file(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    Call the decorated function under the tempfile.NamedTemporaryFile
    context manager. Pass the opened file handle to the decorated function
    """

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        with tempfile.NamedTemporaryFile() as tmpfile:
            return func(*args, tmpfile, **kwargs)

    return wrapper


def mock_anaconda_api_calls(rsps: responses.RequestsMock, module: str, version: str) -> None:
    """Mock anaconda api calls for module"""
    anaconda_api_url = f"https://api.anaconda.org/package/bioconda/{module}"
    anaconda_mock = {
        "latest_version": version.split("--")[0],
        "summary": "",
        "doc_url": "http://test",
        "dev_url": "http://test",
        "files": [{"version": version.split("--")[0]}],
        "license": "MIT",
    }
    rsps.get(anaconda_api_url, json=anaconda_mock, status=200)


def mock_biocontainers_api_calls(rsps: responses.RequestsMock, module: str, version: str) -> None:
    """Mock biocontainers api calls for module"""
    biocontainers_api_url = (
        f"https://api.biocontainers.pro/ga4gh/trs/v2/tools/{module}/versions/{module}-{version.split('--')[0]}"
    )
    biocontainers_mock = {
        "images": [
            {
                "image_type": "Singularity",
                "image_name": f"https://depot.galaxyproject.org/singularity/{module}:{version}",
                "updated": "2021-09-04T00:00:00Z",
            },
            {
                "image_type": "Docker",
                "image_name": f"biocontainers/{module}:{version}",
                "updated": "2021-09-04T00:00:00Z",
            },
        ],
    }
    rsps.get(biocontainers_api_url, json=biocontainers_mock, status=200)


def mock_biotools_api_calls(rsps: responses.RequestsMock, module: str) -> None:
    """Mock biotools api calls for module"""
    biotools_api_url = f"https://bio.tools/api/t/?q={module}&format=json"
    biotools_mock = {
        "list": [{"name": "Bpipe", "biotoolsCURIE": "biotools:bpipe"}],
    }
    rsps.get(biotools_api_url, json=biotools_mock, status=200)


def create_tmp_pipeline(no_git: bool = False) -> Tuple[Path, Path, str, Path]:
    """Create a new Pipeline for testing"""

    tmp_dir = Path(tempfile.TemporaryDirectory().name)
    root_repo_dir = Path(__file__).resolve().parent.parent
    template_dir = root_repo_dir / "nf_core" / "pipeline-template"
    pipeline_name = "testpipeline"
    pipeline_dir = tmp_dir / pipeline_name
    pipeline_dir.mkdir(parents=True)

    nf_core_yml = NFCoreYamlConfig(
        nf_core_version=__version__,
        repository_type="modules",
        org_path="nf-core",
        lint=None,
        template=NFCoreTemplateConfig(
            name="testpipeline",
            author="me",
            description="it is mine",
            org="nf-core",
            version=None,
            force=True,
            is_nfcore=None,
            skip_features=None,
            outdir=None,
        ),
        bump_version=None,
    )
    with open(str(Path(pipeline_dir, ".nf-core.yml")), "w") as fh:
        yaml.dump(nf_core_yml.model_dump(), fh, Dumper=custom_yaml_dumper())

    nf_core.pipelines.create.create.PipelineCreate(
        pipeline_name, "it is mine", "me", no_git=no_git, outdir=pipeline_dir, force=True
    ).init_pipeline()

    # return values to instance variables for later use in test methods
    return tmp_dir, template_dir, pipeline_name, pipeline_dir


def cmp_component(dir1: Path, dir2: Path) -> bool:
    """Compare two versions of the same component"""
    files = ["main.nf", "meta.yml"]
    return all(filecmp.cmp(dir1 / f, dir2 / f, shallow=False) for f in files)



from textwrap import dedent

from textual.app import ComposeResult
from textual.containers import Center
from textual.screen import Screen
from textual.widgets import Button, Footer, Header, Markdown, Static

from nf_core.pipelines.create.utils import add_hide_class
from nf_core.utils import nfcore_logo


class LoggingScreen(Screen):
    """A screen to show the final logs."""

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        yield Markdown(
            dedent(
                """
                # Logging
                """
            )
        )
        yield Static(
            "\n" + "\n".join(nfcore_logo) + "\n",
            id="logo",
        )
        yield Markdown("Creating...")
        yield Center(self.parent.LOG_HANDLER.console)
        yield Center(
            Button("Back", id="back", variant="default", classes="hide"),
            Button("Continue", id="close_screen", variant="success", classes="hide"),
            Button("Continue", id="exit", variant="success", classes="hide"),
            Button("Close App", id="close_app", variant="success", classes="hide"),
            classes="cta",
        )

    def on_screen_resume(self):
        """Hide all buttons as disabled on screen resume."""
        button_ids = ["back", "close_screen", "exit", "close_app"]
        for button in self.query("Button"):
            if button.id in button_ids:
                add_hide_class(self.parent, button.id)

    def on_screen_suspend(self):
        """Clear console on screen suspend."""
        self.parent.LOG_HANDLER.console.clear()



import logging

from nf_core.components.remove import ComponentRemove

log = logging.getLogger(__name__)


class SubworkflowRemove(ComponentRemove):
    def __init__(self, pipeline_dir, remote_url=None, branch=None, no_pull=False):
        super().__init__("subworkflows", pipeline_dir, remote_url=remote_url, branch=branch, no_pull=no_pull)



"""Synchronise a pipeline TEMPLATE branch with the template."""

import json
import logging
import os
import re
import shutil
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Union

import git
import questionary
import requests
import requests.auth
import requests_cache
import rich
import yaml
from git import GitCommandError, InvalidGitRepositoryError

import nf_core
import nf_core.pipelines.create.create
import nf_core.pipelines.list
import nf_core.utils
from nf_core.pipelines.lint_utils import dump_yaml_with_prettier

log = logging.getLogger(__name__)


class SyncExceptionError(Exception):
    """Exception raised when there was an error with TEMPLATE branch synchronisation"""

    pass


class PullRequestExceptionError(Exception):
    """Exception raised when there was an error creating a Pull-Request on GitHub.com"""

    pass


class PipelineSync:
    """Object to hold syncing information and results.

    Args:
        pipeline_dir (str): The path to the Nextflow pipeline root directory
        from_branch (str): The branch to use to fetch config vars. If not set, will use current active branch
        make_pr (bool): Set this to `True` to create a GitHub pull-request with the changes
        gh_username (str): GitHub username
        gh_repo (str): GitHub repository name
        template_yaml_path (str): Path to template.yml file for pipeline creation settings. DEPRECATED
        force_pr (bool): Force the creation of a pull request, even if there are no changes to the template

    Attributes:
        pipeline_dir (str): Path to target pipeline directory
        from_branch (str): Repo branch to use when collecting workflow variables. Default: active branch.
        original_branch (str): Repo branch that was checked out before we started.
        made_changes (bool): Whether making the new template pipeline introduced any changes
        make_pr (bool): Whether to try to automatically make a PR on GitHub.com
        required_config_vars (list): List of nextflow variables required to make template pipeline
        gh_username (str): GitHub username
        gh_repo (str): GitHub repository name
    """

    def __init__(
        self,
        pipeline_dir: Union[str, Path],
        from_branch: Optional[str] = None,
        make_pr: bool = False,
        gh_repo: Optional[str] = None,
        gh_username: Optional[str] = None,
        template_yaml_path: Optional[str] = None,
        force_pr: bool = False,
    ):
        """Initialise syncing object"""

        self.pipeline_dir: Path = Path(pipeline_dir).resolve()
        self.from_branch = from_branch
        self.original_branch = None
        self.original_merge_branch = f"nf-core-template-merge-{nf_core.__version__}"
        self.merge_branch = self.original_merge_branch
        self.made_changes = False
        self.make_pr = make_pr
        self.gh_pr_returned_data: Dict = {}
        self.required_config_vars = ["manifest.name", "manifest.description", "manifest.version", "manifest.author"]
        self.force_pr = force_pr

        self.gh_username = gh_username
        self.gh_repo = gh_repo
        self.pr_url = ""

        self.config_yml_path, self.config_yml = nf_core.utils.load_tools_config(self.pipeline_dir)
        assert self.config_yml_path is not None and self.config_yml is not None  # mypy
        # Throw deprecation warning if template_yaml_path is set
        if template_yaml_path is not None:
            log.warning(
                f"The `template_yaml_path` argument is deprecated. Saving pipeline creation settings in .nf-core.yml instead. Please remove {template_yaml_path} file."
            )
            if getattr(self.config_yml, "template", None) is not None:
                overwrite_template = questionary.confirm(
                    f"A template section already exists in '{self.config_yml_path}'. Do you want to overwrite?",
                    style=nf_core.utils.nfcore_question_style,
                    default=False,
                ).unsafe_ask()
            if overwrite_template or getattr(self.config_yml, "template", None) is None:
                with open(template_yaml_path) as f:
                    self.config_yml.template = yaml.safe_load(f)
                with open(self.config_yml_path, "w") as fh:
                    yaml.safe_dump(self.config_yml.model_dump(exclude_none=True), fh)
                log.info(f"Saved pipeline creation settings to '{self.config_yml_path}'")
                raise SystemExit(
                    f"Please commit your changes and delete the {template_yaml_path} file. Then run the sync command again."
                )

        # Set up the API auth if supplied on the command line
        self.gh_api = nf_core.utils.gh_api
        self.gh_api.lazy_init()
        if self.gh_username and "GITHUB_AUTH_TOKEN" in os.environ:
            log.debug(f"Authenticating sync as {self.gh_username}")
            self.gh_api.setup_github_auth(
                requests.auth.HTTPBasicAuth(self.gh_username, os.environ["GITHUB_AUTH_TOKEN"])
            )

    def sync(self) -> None:
        """Find workflow attributes, create a new template pipeline on TEMPLATE"""

        # Clear requests_cache so that we don't get stale API responses
        requests_cache.clear()

        log.info(f"Pipeline directory: {self.pipeline_dir}")
        if self.from_branch:
            log.info(f"Using branch '{self.from_branch}' to fetch workflow variables")
        if self.make_pr:
            log.info("Will attempt to automatically create a pull request")

        self.inspect_sync_dir()
        self.get_wf_config()
        self.checkout_template_branch()
        self.delete_template_branch_files()
        self.make_template_pipeline()
        self.commit_template_changes()

        if not self.made_changes and self.force_pr:
            log.info("No changes made to TEMPLATE, but PR forced")
            self.made_changes = True

        # Push and make a pull request if we've been asked to
        if self.made_changes and self.make_pr or self.force_pr:
            try:
                # Check that we have an API auth token
                if os.environ.get("GITHUB_AUTH_TOKEN", "") == "":
                    raise PullRequestExceptionError("GITHUB_AUTH_TOKEN not set!")

                # Check that we know the github username and repo name
                if self.gh_username is None and self.gh_repo is None:
                    raise PullRequestExceptionError("Could not find GitHub username and repo name")

                self.push_template_branch()
                self.create_merge_base_branch()
                self.push_merge_branch()
                self.make_pull_request()
                self.close_open_template_merge_prs()
            except PullRequestExceptionError as e:
                self.reset_target_dir()
                raise PullRequestExceptionError(e)

        self.reset_target_dir()

        if not self.made_changes:
            log.info("No changes made to TEMPLATE - sync complete")
        elif not self.make_pr:
            log.info(
                f"Now try to merge the updates in to your pipeline:\n  cd {self.pipeline_dir}\n  git merge TEMPLATE"
            )

    def inspect_sync_dir(self):
        """Takes a look at the target directory for syncing. Checks that it's a git repo
        and makes sure that there are no uncommitted changes.
        """
        # Check that the pipeline_dir is a git repo
        try:
            self.repo = git.Repo(self.pipeline_dir)
        except InvalidGitRepositoryError:
            raise SyncExceptionError(f"'{self.pipeline_dir}' does not appear to be a git repository")

        # get current branch so we can switch back later
        self.original_branch = self.repo.active_branch.name
        log.info(f"Original pipeline repository branch is '{self.original_branch}'")

        # Check to see if there are uncommitted changes on current branch
        if self.repo.is_dirty(untracked_files=True):
            raise SyncExceptionError(
                "Uncommitted changes found in pipeline directory!\nPlease commit these before running nf-core pipelines sync"
            )

    def get_wf_config(self):
        """Check out the target branch if requested and fetch the nextflow config.
        Check that we have the required config variables.
        """
        # Try to check out target branch (eg. `origin/dev`)
        try:
            if self.from_branch and self.repo.active_branch.name != self.from_branch:
                log.info(f"Checking out workflow branch '{self.from_branch}'")
                self.repo.git.checkout(self.from_branch)
        except GitCommandError:
            raise SyncExceptionError(f"Branch `{self.from_branch}` not found!")

        # If not specified, get the name of the active branch
        if not self.from_branch:
            try:
                self.from_branch = self.repo.active_branch.name
            except GitCommandError as e:
                log.error(f"Could not find active repo branch: {e}")

        # Fetch workflow variables
        log.debug("Fetching workflow config variables")
        self.wf_config = nf_core.utils.fetch_wf_config(Path(self.pipeline_dir))

        # Check that we have the required variables
        for rvar in self.required_config_vars:
            if rvar not in self.wf_config:
                raise SyncExceptionError(f"Workflow config variable `{rvar}` not found!")

    def checkout_template_branch(self):
        """
        Try to check out the origin/TEMPLATE in a new TEMPLATE branch.
        If this fails, try to check out an existing local TEMPLATE branch.
        """
        # Try to check out the `TEMPLATE` branch
        try:
            self.repo.git.checkout("origin/TEMPLATE", b="TEMPLATE")
        except GitCommandError:
            # Try to check out an existing local branch called TEMPLATE
            try:
                self.repo.git.checkout("TEMPLATE")
            except GitCommandError:
                raise SyncExceptionError("Could not check out branch 'origin/TEMPLATE' or 'TEMPLATE'")

    def delete_template_branch_files(self):
        """
        Delete all files in the TEMPLATE branch
        """
        # Delete everything
        log.info("Deleting all files in 'TEMPLATE' branch")
        for the_file in os.listdir(self.pipeline_dir):
            if the_file == ".git":
                continue
            file_path = os.path.join(self.pipeline_dir, the_file)
            log.debug(f"Deleting {file_path}")
            try:
                if os.path.isfile(file_path):
                    os.unlink(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)
            except Exception as e:
                raise SyncExceptionError(e)

    def make_template_pipeline(self):
        """
        Delete all files and make a fresh template using the workflow variables
        """
        log.info("Making a new template pipeline using pipeline variables")

        # Only show error messages from pipeline creation
        logging.getLogger("nf_core.pipelines.create").setLevel(logging.ERROR)
        assert self.config_yml_path is not None
        assert self.config_yml is not None

        # Re-write the template yaml info from .nf-core.yml config
        if self.config_yml.template is not None:
            # Set force true in config to overwrite existing files

            self.config_yml.template.force = True
            with open(self.config_yml_path, "w") as config_path:
                yaml.safe_dump(self.config_yml.model_dump(exclude_none=True), config_path)

        try:
            pipeline_create_obj = nf_core.pipelines.create.create.PipelineCreate(
                outdir=str(self.pipeline_dir),
                from_config_file=True,
                no_git=True,
                force=True,
            )
            pipeline_create_obj.init_pipeline()

            # set force to false to avoid overwriting files in the future
            if self.config_yml.template is not None:
                self.config_yml.template = pipeline_create_obj.config
                # Set force true in config to overwrite existing files
                self.config_yml.template.force = False
                # Set outdir as the current directory to avoid local info leaking
                self.config_yml.template.outdir = "."
                # Update nf-core version
                self.config_yml.nf_core_version = nf_core.__version__
                dump_yaml_with_prettier(self.config_yml_path, self.config_yml.model_dump(exclude_none=True))

        except Exception as err:
            # Reset to where you were to prevent git getting messed up.
            self.repo.git.reset("--hard")
            raise SyncExceptionError(f"Failed to rebuild pipeline from template with error:\n{err}")

    def commit_template_changes(self):
        """If we have any changes with the new template files, make a git commit"""
        # Check that we have something to commit
        if not self.repo.is_dirty(untracked_files=True):
            log.info("Template contains no changes - no new commit created")
            return False
        # Commit changes
        try:
            self.repo.git.add(A=True)
            self.repo.index.commit(f"Template update for nf-core/tools version {nf_core.__version__}")
            self.made_changes = True
            log.info("Committed changes to 'TEMPLATE' branch")
        except Exception as e:
            raise SyncExceptionError(f"Could not commit changes to TEMPLATE:\n{e}")
        return True

    def push_template_branch(self):
        """If we made any changes, push the TEMPLATE branch to the default remote
        and try to make a PR. If we don't have the auth token, try to figure out a URL
        for the PR and print this to the console.
        """
        log.info(f"Pushing TEMPLATE branch to remote: '{os.path.basename(self.pipeline_dir)}'")
        try:
            self.repo.git.push()
        except GitCommandError as e:
            raise PullRequestExceptionError(f"Could not push TEMPLATE branch:\n  {e}")

    def create_merge_base_branch(self):
        """Create a new branch from the updated TEMPLATE branch
        This branch will then be used to create the PR
        """
        # Check if branch exists already
        branch_list = [b.name for b in self.repo.branches]
        if self.merge_branch in branch_list:
            merge_branch_format = re.compile(rf"{self.original_merge_branch}-(\d+)")
            max_branch = max(
                [1]
                + [
                    int(merge_branch_format.match(branch).groups()[0])
                    for branch in branch_list
                    if merge_branch_format.match(branch)
                ]
            )
            new_branch = f"{self.original_merge_branch}-{max_branch+1}"
            log.info(f"Branch already existed: '{self.merge_branch}', creating branch '{new_branch}' instead.")
            self.merge_branch = new_branch

        # Create new branch and checkout
        log.info(f"Checking out merge base branch '{self.merge_branch}'")
        try:
            self.repo.create_head(self.merge_branch)
        except GitCommandError as e:
            raise SyncExceptionError(f"Could not create new branch '{self.merge_branch}'\n{e}")

    def push_merge_branch(self):
        """Push the newly created merge branch to the remote repository"""
        log.info(f"Pushing '{self.merge_branch}' branch to remote")
        try:
            origin = self.repo.remote()
            origin.push(self.merge_branch)
        except GitCommandError as e:
            raise PullRequestExceptionError(f"Could not push branch '{self.merge_branch}':\n  {e}")

    def make_pull_request(self):
        """Create a pull request to a base branch (default: dev),
        from a head branch (default: TEMPLATE)

        Returns: An instance of class requests.Response
        """
        log.info("Submitting a pull request via the GitHub API")

        pr_title = f"Important! Template update for nf-core/tools v{nf_core.__version__}"
        pr_body_text = (
            "Version `{tag}` of [nf-core/tools](https://github.com/nf-core/tools) has just been released with updates to the nf-core template. "
            "This automated pull-request attempts to apply the relevant updates to this pipeline.\n\n"
            "Please make sure to merge this pull-request as soon as possible, "
            f"resolving any merge conflicts in the `{self.merge_branch}` branch (or your own fork, if you prefer). "
            "Once complete, make a new minor release of your pipeline.\n\n"
            "For instructions on how to merge this PR, please see "
            "[https://nf-co.re/docs/contributing/sync/](https://nf-co.re/docs/contributing/sync/#merging-automated-prs).\n\n"
            "For more information about this release of [nf-core/tools](https://github.com/nf-core/tools), "
            "please see the `v{tag}` [release page](https://github.com/nf-core/tools/releases/tag/{tag})."
        ).format(tag=nf_core.__version__)

        # Make new pull-request
        stderr = rich.console.Console(stderr=True, force_terminal=nf_core.utils.rich_force_colors())
        with self.gh_api.cache_disabled():
            try:
                r = self.gh_api.request_retry(
                    f"https://api.github.com/repos/{self.gh_repo}/pulls",
                    post_data={
                        "title": pr_title,
                        "body": pr_body_text,
                        "maintainer_can_modify": True,
                        "head": self.merge_branch,
                        "base": self.from_branch,
                    },
                )
            except Exception as e:
                stderr.print_exception()
                raise PullRequestExceptionError(f"Something went badly wrong - {e}")
            else:
                self.gh_pr_returned_data = r.json()
                self.pr_url = self.gh_pr_returned_data["html_url"]
                log.debug(f"GitHub API PR worked, return code {r.status_code}")
                log.info(f"GitHub PR created: {self.gh_pr_returned_data['html_url']}")

    def close_open_template_merge_prs(self):
        """Get all template merging branches (starting with 'nf-core-template-merge-')
        and check for any open PRs from these branches to the self.from_branch
        If open PRs are found, add a comment and close them
        """
        log.info("Checking for open PRs from template merge branches")

        # Look for existing pull-requests
        list_prs_url = f"https://api.github.com/repos/{self.gh_repo}/pulls"
        with self.gh_api.cache_disabled():
            list_prs_request = self.gh_api.get(list_prs_url)

        list_prs_json, list_prs_pp = self._parse_json_response(list_prs_request)

        log.debug(f"GitHub API listing existing PRs:\n{list_prs_url}\n{list_prs_pp}")
        if list_prs_request.status_code != 200:
            log.warning(f"Could not list open PRs ('{list_prs_request.status_code}')\n{list_prs_url}\n{list_prs_pp}")
            return False

        for pr in list_prs_json:
            if isinstance(pr, int):
                log.debug(f"Incorrect PR format: {pr}")
            else:
                log.debug(f"Looking at PR from '{pr['head']['ref']}': {pr['html_url']}")
                # Ignore closed PRs
                if pr["state"] != "open":
                    log.debug(f"Ignoring PR as state not open ({pr['state']}): {pr['html_url']}")
                    continue

                # Don't close the new PR that we just opened
                if pr["head"]["ref"] == self.merge_branch:
                    continue

                # PR is from an automated branch and goes to our target base
                if pr["head"]["ref"].startswith("nf-core-template-merge-") and pr["base"]["ref"] == self.from_branch:
                    self.close_open_pr(pr)

    def close_open_pr(self, pr) -> bool:
        """Given a PR API response, add a comment and close."""
        log.debug(f"Attempting to close PR: '{pr['html_url']}'")

        # Make a new comment explaining why the PR is being closed
        comment_text = (
            f"Version `{nf_core.__version__}` of the [nf-core/tools](https://github.com/nf-core/tools) pipeline template has just been released. "
            f"This pull-request is now outdated and has been closed in favour of {self.pr_url}\n\n"
            f"Please use {self.pr_url} to merge in the new changes from the nf-core template as soon as possible."
        )
        with self.gh_api.cache_disabled():
            self.gh_api.post(url=pr["comments_url"], data=json.dumps({"body": comment_text}))

        # Update the PR status to be closed
        with self.gh_api.cache_disabled():
            pr_request = self.gh_api.patch(url=pr["url"], data=json.dumps({"state": "closed"}))

        pr_request_json, pr_request_pp = self._parse_json_response(pr_request)

        # PR update worked
        if pr_request.status_code == 200:
            log.debug(f"GitHub API PR-update worked:\n{pr_request_pp}")
            log.info(
                f"Closed GitHub PR from '{pr['head']['ref']}' to '{pr['base']['ref']}': {pr_request_json['html_url']}"
            )
            return True
        # Something went wrong
        else:
            log.warning(f"Could not close PR ('{pr_request.status_code}'):\n{pr['url']}\n{pr_request_pp}")
            return False

    @staticmethod
    def _parse_json_response(response) -> Tuple[Any, str]:
        """Helper method to parse JSON response and create pretty-printed string.

        Args:
            response: requests.Response object

        Returns:
            Tuple of (parsed_json, pretty_printed_str)
        """
        try:
            json_data = json.loads(response.content)
            return json_data, json.dumps(json_data, indent=4)
        except Exception:
            return response.content, str(response.content)

    def reset_target_dir(self):
        """
        Reset the target pipeline directory. Check out the original branch.
        """
        log.info(f"Checking out original branch: '{self.original_branch}'")
        try:
            self.repo.git.checkout(self.original_branch)
        except GitCommandError as e:
            raise SyncExceptionError(f"Could not reset to original branch `{self.original_branch}`:\n{e}")



import logging
import os
import shutil
import tempfile
from pathlib import Path

import questionary

import nf_core.utils
from nf_core.components.components_command import ComponentCommand
from nf_core.components.components_differ import ComponentsDiffer
from nf_core.modules.modules_json import ModulesJson

log = logging.getLogger(__name__)


class ComponentPatch(ComponentCommand):
    def __init__(self, pipeline_dir, component_type, remote_url=None, branch=None, no_pull=False, installed_by=None):
        super().__init__(component_type, pipeline_dir, remote_url, branch, no_pull)

        self.modules_json = ModulesJson(pipeline_dir)

    def _parameter_checks(self, component):
        """Checks the compatibility of the supplied parameters.

        Raises:
            UserWarning: if any checks fail.
        """
        if not self.has_valid_directory():
            raise UserWarning("The command was not run in a valid pipeline directory.")

        components = self.modules_json.get_all_components(self.component_type).get(self.modules_repo.remote_url)
        if components is None:
            raise UserWarning(
                f"No {self.component_type[:-1]}s found in the 'modules.json' file for the remote '{self.modules_repo.remote_url}'"
            )
        component_names = [component for _, component in components]

        if component is not None and component not in component_names:
            component_dir = [d for d, m in components if m == component][0]
            raise UserWarning(
                f"{self.component_type[:-1].title()} '{Path(self.component_type, component_dir, component)}' does not exist in the pipeline"
            )

    def patch(self, component=None):
        # Check modules directory structure
        self.check_modules_structure()

        # Verify that 'modules.json' is consistent with the installed modules
        self.modules_json.check_up_to_date()
        self._parameter_checks(component)
        components = self.modules_json.get_all_components(self.component_type).get(self.modules_repo.remote_url)

        if component is None:
            choices = [
                component if directory == self.modules_repo.repo_path else f"{directory}/{component}"
                for directory, component in components
            ]
            component = questionary.autocomplete(
                f"{self.component_type[:-1].title()} name:",
                choices=sorted(choices),
                style=nf_core.utils.nfcore_question_style,
            ).unsafe_ask()
        component_dir = [dir for dir, m in components if m == component][0]
        component_fullname = str(Path(self.component_type, self.modules_repo.repo_path, component))

        # Verify that the component has an entry in the modules.json file
        if not self.modules_json.component_present(
            component, self.modules_repo.remote_url, component_dir, self.component_type
        ):
            raise UserWarning(
                f"The '{component_fullname}' {self.component_type[:-1]} does not have an entry in the 'modules.json' file. Cannot compute patch"
            )

        component_version = self.modules_json.get_component_version(
            self.component_type, component, self.modules_repo.remote_url, self.modules_repo.repo_path
        )
        if component_version is None:
            raise UserWarning(
                f"The '{component_fullname}' {self.component_type[:-1]} does not have a valid version in the 'modules.json' file. Cannot compute patch"
            )
        # Get the component branch and reset it in the ModulesRepo object
        component_branch = self.modules_json.get_component_branch(
            self.component_type, component, self.modules_repo.remote_url, component_dir
        )
        if component_branch != self.modules_repo.branch:
            self.modules_repo.setup_branch(component_branch)

        # Set the diff filename based on the module name
        patch_filename = f"{component.replace('/', '-')}.diff"
        component_relpath = Path(self.component_type, component_dir, component)
        patch_relpath = Path(component_relpath, patch_filename)
        component_current_dir = Path(self.directory, component_relpath)
        patch_path = Path(self.directory, patch_relpath)

        if patch_path.exists():
            remove = questionary.confirm(
                f"Patch exists for {self.component_type[:-1]} '{component_fullname}'. Do you want to regenerate it?",
                style=nf_core.utils.nfcore_question_style,
            ).unsafe_ask()
            if remove:
                os.remove(patch_path)
            else:
                return

        # Create a temporary directory for storing the unchanged version of the module
        install_dir = tempfile.mkdtemp()
        component_install_dir = Path(install_dir, component)
        if not self.install_component_files(component, component_version, self.modules_repo, install_dir):
            raise UserWarning(
                f"Failed to install files of {self.component_type[:-1]} '{component}' from remote ({self.modules_repo.remote_url})."
            )

        # Write the patch to a temporary location (otherwise it is printed to the screen later)
        patch_temp_path = tempfile.mktemp()
        try:
            ComponentsDiffer.write_diff_file(
                patch_temp_path,
                component,
                self.modules_repo.repo_path,
                component_install_dir,
                component_current_dir,
                for_git=False,
                dsp_from_dir=component_relpath,
                dsp_to_dir=component_relpath,
            )
            log.debug(f"Patch file wrote to a temporary directory {patch_temp_path}")
        except UserWarning:
            raise UserWarning(f"{self.component_type[:-1]} '{component_fullname}' is unchanged. No patch to compute")

        # Write changes to modules.json
        self.modules_json.add_patch_entry(
            self.component_type, component, self.modules_repo.remote_url, component_dir, patch_relpath
        )
        log.debug(f"Wrote patch path for {self.component_type[:-1]} {component} to modules.json")

        # Show the changes made to the module
        ComponentsDiffer.print_diff(
            component,
            self.modules_repo.repo_path,
            component_install_dir,
            component_current_dir,
            dsp_from_dir=component_current_dir,
            dsp_to_dir=component_current_dir,
        )

        # Finally move the created patch file to its final location
        shutil.move(patch_temp_path, patch_path)
        log.info(f"Patch file of '{component_fullname}' written to '{patch_path}'")

    def remove(self, component):
        # Check modules directory structure
        self.check_modules_structure()

        self.modules_json.check_up_to_date()
        self._parameter_checks(component)
        components = self.modules_json.get_all_components(self.component_type).get(self.modules_repo.remote_url)

        if component is None:
            choices = [
                component if directory == self.modules_repo.repo_path else f"{directory}/{component}"
                for directory, component in components
            ]
            component = questionary.autocomplete(
                f"{self.component_type[:-1]} name:",
                choices,
                style=nf_core.utils.nfcore_question_style,
            ).unsafe_ask()
        component_dir = [dir for dir, m in components if m == component][0]
        component_fullname = str(Path(self.component_type, component_dir, component))

        # Verify that the component has an entry in the modules.json file
        if not self.modules_json.component_present(
            component, self.modules_repo.remote_url, component_dir, self.component_type
        ):
            raise UserWarning(
                f"The '{component_fullname}' {self.component_type[:-1]} does not have an entry in the 'modules.json' file. Cannot compute patch"
            )

        component_version = self.modules_json.get_component_version(
            self.component_type, component, self.modules_repo.remote_url, self.modules_repo.repo_path
        )
        if component_version is None:
            raise UserWarning(
                f"The '{component_fullname}' {self.component_type[:-1]} does not have a valid version in the 'modules.json' file. Cannot compute patch"
            )
        # Get the module branch and reset it in the ModulesRepo object
        component_branch = self.modules_json.get_component_branch(
            self.component_type, component, self.modules_repo.remote_url, component_dir
        )
        if component_branch != self.modules_repo.branch:
            self.modules_repo.setup_branch(component_branch)

        # Set the diff filename based on the component name
        patch_filename = f"{component.replace('/', '-')}.diff"
        component_relpath = Path(self.component_type, component_dir, component)
        patch_relpath = Path(component_relpath, patch_filename)
        patch_path = Path(self.directory, patch_relpath)
        component_path = Path(self.directory, component_relpath)

        if patch_path.exists():
            remove = questionary.confirm(
                f"Patch exists for {self.component_type[:-1]} '{component_fullname}'. Are you sure you want to remove?",
                style=nf_core.utils.nfcore_question_style,
            ).unsafe_ask()
            if not remove:
                return

        # Try to apply the patch in reverse and move resulting files to module dir
        temp_component_dir = self.modules_json.try_apply_patch_reverse(
            self.component_type, component, self.modules_repo.repo_path, patch_relpath, component_path
        )
        try:
            for file in Path(temp_component_dir).glob("*"):
                file.rename(component_path.joinpath(file.name))
            os.rmdir(temp_component_dir)
        except Exception as err:
            raise UserWarning(f"There was a problem reverting the patched file: {err}")

        log.info(f"Patch for {component} reverted!")
        # Remove patch file if we could revert the patch
        patch_path.unlink()
        # Write changes to module.json
        self.modules_json.remove_patch_entry(component, self.modules_repo.remote_url, component_dir)

        if not all(
            self.modules_repo.component_files_identical(
                component, component_path, component_version, self.component_type
            ).values()
        ):
            log.error(
                f"Module files do not appear to match the remote for the commit sha in the 'module.json': {component_version}\n"
                f"Recommend reinstalling with 'nf-core modules install --force --sha {component_version} {component}' "
            )



"""Tests covering the sync command"""

import json
import os
from pathlib import Path
from typing import Dict, List, Union
from unittest import mock

import git
import pytest
import yaml

import nf_core.pipelines.create.create
import nf_core.pipelines.sync
from nf_core.utils import NFCoreYamlConfig

from ..test_pipelines import TestPipelines
from ..utils import with_temporary_folder


class MockResponse:
    def __init__(self, data: Union[Dict, List[Dict]], status_code: int, url: str):
        self.url: str = url
        self.status_code: int = status_code
        self.from_cache: bool = False
        self.reason: str = "Mocked response"
        self.data: Union[Dict, List[Dict]] = data
        self.content: str = json.dumps(data)
        self.headers: Dict[str, str] = {"content-encoding": "test", "connection": "fake"}

    def json(self):
        return self.data


def mocked_requests_get(url) -> MockResponse:
    """Helper function to emulate POST requests responses from the web"""

    url_template = "https://api.github.com/repos/{}/response/"
    if url == Path(url_template.format("no_existing_pr"), "pulls?head=TEMPLATE&base=None"):
        return MockResponse([], 200, url)
    if url == Path(url_template.format("list_prs"), "pulls"):
        response_data = [
            {
                "state": "closed",
                "head": {"ref": "nf-core-template-merge-2"},
                "base": {"ref": "main"},
                "html_url": "pr_url",
            }
        ] + [
            {
                "state": "open",
                "head": {"ref": f"nf-core-template-merge-{branch_no}"},
                "base": {"ref": "main"},
                "html_url": "pr_url",
            }
            for branch_no in range(3, 7)
        ]
        return MockResponse(response_data, 200, url)
    if url == "https://nf-co.re/pipelines.json":
        return MockResponse({"remote_workflows": [{"name": "testpipeline", "topics": ["test", "pipeline"]}]}, 200, url)

    return MockResponse([{"html_url": url}], 404, url)


def mocked_requests_patch(url: str, data: str, **kwargs) -> MockResponse:
    """Helper function to emulate POST requests responses from the web"""

    if url == "url_to_update_pr":
        return MockResponse({"html_url": "great_success"}, 200, url)
    # convert data to dict
    response = json.loads(data)
    response["patch_url"] = url
    return MockResponse(response, 404, url)


def mocked_requests_post(url, **kwargs):
    """Helper function to emulate POST requests responses from the web"""

    if url == "https://api.github.com/repos/no_existing_pr/response/pulls":
        return MockResponse({"html_url": "great_success"}, 201, url)

    return MockResponse({}, 404, url)


class TestModules(TestPipelines):
    """Class for modules tests"""

    def setUp(self):
        super().setUp()
        self.remote_path = Path(self.tmp_dir, "remote_repo")
        self.remote_repo = git.Repo.init(self.remote_path, bare=True)

        if self.remote_repo.active_branch.name != "master":
            self.remote_repo.active_branch.rename("master")

    @with_temporary_folder
    def test_inspect_sync_dir_notgit(self, tmp_dir: str):
        """Try syncing an empty directory"""
        nf_core_yml_path = Path(tmp_dir, ".nf-core.yml")
        nf_core_yml = NFCoreYamlConfig(repository_type="pipeline")

        with open(nf_core_yml_path, "w") as fh:
            yaml.dump(nf_core_yml.model_dump(), fh)

        psync = nf_core.pipelines.sync.PipelineSync(tmp_dir)
        with pytest.raises(nf_core.pipelines.sync.SyncExceptionError) as exc_info:
            psync.inspect_sync_dir()
        assert "does not appear to be a git repository" in exc_info.value.args[0]

    def test_inspect_sync_dir_dirty(self):
        """Try syncing a pipeline with uncommitted changes"""
        # Add an empty file, uncommitted
        test_fn = Path(self.pipeline_dir) / "uncommitted"
        test_fn.touch()
        # Try to sync, check we halt with the right error
        psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir)
        try:
            with pytest.raises(nf_core.pipelines.sync.SyncExceptionError) as exc_info:
                psync.inspect_sync_dir()
            assert exc_info.value.args[0].startswith("Uncommitted changes found in pipeline directory!")
        finally:
            os.remove(test_fn)

    def test_get_wf_config_no_branch(self):
        """Try getting a workflow config when the branch doesn't exist"""
        # Try to sync, check we halt with the right error
        psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir, from_branch="foo")
        with pytest.raises(nf_core.pipelines.sync.SyncExceptionError) as exc_info:
            psync.inspect_sync_dir()
            psync.get_wf_config()
        assert exc_info.value.args[0] == "Branch `foo` not found!"

    def test_get_wf_config_missing_required_config(self):
        """Try getting a workflow config, then make it miss a required config option"""
        # Try to sync, check we halt with the right error
        psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir)
        psync.required_config_vars = ["fakethisdoesnotexist"]
        with pytest.raises(nf_core.pipelines.sync.SyncExceptionError) as exc_info:
            psync.inspect_sync_dir()
            psync.get_wf_config()
        # Check that we did actually get some config back
        assert psync.wf_config["params.validate_params"] == "true"
        # Check that we raised because of the missing fake config var
        assert exc_info.value.args[0] == "Workflow config variable `fakethisdoesnotexist` not found!"

    def test_checkout_template_branch(self):
        """Try checking out the TEMPLATE branch of the pipeline"""
        psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir)
        psync.inspect_sync_dir()
        psync.get_wf_config()
        psync.checkout_template_branch()

    def test_checkout_template_branch_no_template(self):
        """Try checking out the TEMPLATE branch of the pipeline when it does not exist"""
        psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir)
        psync.inspect_sync_dir()
        psync.get_wf_config()

        psync.repo.delete_head("TEMPLATE")

        with pytest.raises(nf_core.pipelines.sync.SyncExceptionError) as exc_info:
            psync.checkout_template_branch()
        assert exc_info.value.args[0] == "Could not check out branch 'origin/TEMPLATE' or 'TEMPLATE'"

    def test_delete_template_branch_files(self):
        """Confirm that we can delete all files in the TEMPLATE branch"""
        psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir)
        psync.inspect_sync_dir()
        psync.get_wf_config()
        psync.checkout_template_branch()
        psync.delete_template_branch_files()
        assert os.listdir(self.pipeline_dir) == [".git"]

    def test_create_template_pipeline(self):
        """Confirm that we can delete all files in the TEMPLATE branch"""
        # First, delete all the files
        psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir)
        psync.inspect_sync_dir()
        psync.get_wf_config()
        psync.checkout_template_branch()
        psync.delete_template_branch_files()
        assert os.listdir(self.pipeline_dir) == [".git"]
        # Now create the new template
        psync.make_template_pipeline()
        assert "main.nf" in os.listdir(self.pipeline_dir)
        assert "nextflow.config" in os.listdir(self.pipeline_dir)

    def test_commit_template_changes_nochanges(self):
        """Try to commit the TEMPLATE branch, but no changes were made"""
        # Check out the TEMPLATE branch but skip making the new template etc.
        psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir)
        psync.inspect_sync_dir()
        psync.get_wf_config()
        psync.checkout_template_branch()
        # Function returns False if no changes were made
        assert psync.commit_template_changes() is False

    def test_commit_template_changes_changes(self):
        """Try to commit the TEMPLATE branch, but no changes were made"""
        # Check out the TEMPLATE branch but skip making the new template etc.
        psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir)
        psync.inspect_sync_dir()
        psync.get_wf_config()
        psync.checkout_template_branch()
        # Add an empty file, uncommitted
        test_fn = Path(self.pipeline_dir) / "uncommitted"
        test_fn.touch()
        # Check that we have uncommitted changes
        assert psync.repo.is_dirty(untracked_files=True) is True
        # Function returns True if no changes were made
        assert psync.commit_template_changes() is True
        # Check that we don't have any uncommitted changes
        assert psync.repo.is_dirty(untracked_files=True) is False

    def test_push_template_branch_error(self):
        """Try pushing the changes, but without a remote (should fail)"""
        # Check out the TEMPLATE branch but skip making the new template etc.
        psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir)
        psync.inspect_sync_dir()
        psync.get_wf_config()
        psync.checkout_template_branch()
        # Add an empty file and commit it
        test_fn = Path(self.pipeline_dir) / "uncommitted"
        test_fn.touch()
        psync.commit_template_changes()
        # Try to push changes
        with pytest.raises(nf_core.pipelines.sync.PullRequestExceptionError) as exc_info:
            psync.push_template_branch()
        assert exc_info.value.args[0].startswith("Could not push TEMPLATE branch")

    def test_create_merge_base_branch(self):
        """Try creating a merge base branch"""
        psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir)
        psync.inspect_sync_dir()
        psync.get_wf_config()

        psync.create_merge_base_branch()

        assert psync.merge_branch in psync.repo.branches

    def test_create_merge_base_branch_thrice(self):
        """Try creating a merge base branch thrice

        This is needed because the first time this function is called, the
        merge branch does not exist yet (it is only created at the end of the
        create_merge_base_branch function) and the if-statement is ignored.
        Also, the second time this function is called, the existing merge
        branch only has the base format, i.e. without the -{branch_no} at the
        end, so it is needed to call it a third time to make sure this is
        picked up.
        """
        psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir)
        psync.inspect_sync_dir()
        psync.get_wf_config()

        for _ in range(3):
            psync.create_merge_base_branch()

        assert psync.merge_branch in psync.repo.branches
        for branch_no in [2, 3]:
            assert f"{psync.original_merge_branch}-{branch_no}" in psync.repo.branches

    def test_push_merge_branch(self):
        """Try pushing merge branch"""
        psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir)
        psync.inspect_sync_dir()
        psync.get_wf_config()
        psync.repo.create_remote("origin", self.remote_path)

        psync.create_merge_base_branch()
        psync.push_merge_branch()

        assert psync.merge_branch in [b.name for b in self.remote_repo.branches]

    def test_push_merge_branch_without_create_branch(self):
        """Try pushing merge branch without creating first"""
        psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir)
        psync.inspect_sync_dir()
        psync.get_wf_config()
        psync.repo.create_remote("origin", self.remote_path)

        with pytest.raises(nf_core.pipelines.sync.PullRequestExceptionError) as exc_info:
            psync.push_merge_branch()
        assert exc_info.value.args[0].startswith(f"Could not push branch '{psync.merge_branch}'")

    @mock.patch("nf_core.utils.gh_api.get", side_effect=mocked_requests_get)
    @mock.patch("nf_core.utils.gh_api.post", side_effect=mocked_requests_post)
    def test_make_pull_request_success(self, mock_post, mock_get):
        """Try making a PR - successful response"""
        psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir)
        psync.gh_api.get = mock_get
        psync.gh_api.post = mock_post
        psync.gh_username = "no_existing_pr"
        psync.gh_repo = "no_existing_pr/response"
        os.environ["GITHUB_AUTH_TOKEN"] = "test"
        psync.make_pull_request()
        assert psync.gh_pr_returned_data["html_url"] == "great_success"

    @mock.patch("nf_core.utils.gh_api.get", side_effect=mocked_requests_get)
    @mock.patch("nf_core.utils.gh_api.post", side_effect=mocked_requests_post)
    def test_make_pull_request_bad_response(self, mock_post, mock_get):
        """Try making a PR and getting a 404 error"""
        psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir)
        psync.gh_api.get = mock_get
        psync.gh_api.post = mock_post
        psync.gh_username = "bad_url"
        psync.gh_repo = "bad_url/response"
        os.environ["GITHUB_AUTH_TOKEN"] = "test"
        with pytest.raises(nf_core.pipelines.sync.PullRequestExceptionError) as exc_info:
            psync.make_pull_request()
        assert exc_info.value.args[0].startswith(
            "Something went badly wrong - GitHub API PR failed - got return code 404"
        )

    @mock.patch("nf_core.utils.gh_api.get", side_effect=mocked_requests_get)
    def test_close_open_template_merge_prs(self, mock_get):
        """Try closing all open prs"""
        psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir)
        psync.inspect_sync_dir()
        psync.get_wf_config()
        psync.gh_api.get = mock_get
        psync.gh_username = "list_prs"
        psync.gh_repo = "list_prs/response"
        os.environ["GITHUB_AUTH_TOKEN"] = "test"

        with mock.patch("nf_core.pipelines.sync.PipelineSync.close_open_pr") as mock_close_open_pr:
            psync.close_open_template_merge_prs()

            prs = mock_get(f"https://api.github.com/repos/{psync.gh_repo}/pulls").data
            for pr in prs:
                if pr.get("state", None) == "open":
                    mock_close_open_pr.assert_any_call(pr)

    @mock.patch("nf_core.utils.gh_api.post", side_effect=mocked_requests_post)
    @mock.patch("nf_core.utils.gh_api.patch", side_effect=mocked_requests_patch)
    def test_close_open_pr(self, mock_patch, mock_post) -> None:
        psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir)
        psync.inspect_sync_dir()
        psync.get_wf_config()
        psync.gh_api.post = mock_post
        psync.gh_api.patch = mock_patch
        psync.gh_username = "bad_url"
        psync.gh_repo = "bad_url/response"
        os.environ["GITHUB_AUTH_TOKEN"] = "test"
        pr: Dict[str, Union[str, Dict[str, str]]] = {
            "state": "open",
            "head": {"ref": "nf-core-template-merge-3"},
            "base": {"ref": "main"},
            "html_url": "pr_html_url",
            "url": "url_to_update_pr",
            "comments_url": "pr_comments_url",
        }

        assert psync.close_open_pr(pr)
        mock_patch.assert_called_once_with(url="url_to_update_pr", data='{"state": "closed"}')

    @mock.patch("nf_core.utils.gh_api.post", side_effect=mocked_requests_post)
    @mock.patch("nf_core.utils.gh_api.patch", side_effect=mocked_requests_patch)
    def test_close_open_pr_fail(self, mock_patch, mock_post):
        psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir)
        psync.inspect_sync_dir()
        psync.get_wf_config()
        psync.gh_api.post = mock_post
        psync.gh_api.patch = mock_patch
        psync.gh_username = "bad_url"
        psync.gh_repo = "bad_url/response"
        os.environ["GITHUB_AUTH_TOKEN"] = "test"
        pr = {
            "state": "open",
            "head": {"ref": "nf-core-template-merge-3"},
            "base": {"ref": "main"},
            "html_url": "pr_html_url",
            "url": "bad_url_to_update_pr",
            "comments_url": "pr_comments_url",
        }

        assert not psync.close_open_pr(pr)
        mock_patch.assert_called_once_with(url="bad_url_to_update_pr", data='{"state": "closed"}')

    def test_reset_target_dir(self):
        """Try resetting target pipeline directory"""
        psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir)
        psync.inspect_sync_dir()
        psync.get_wf_config()

        psync.repo.git.checkout("dev")

        psync.reset_target_dir()

        assert psync.repo.heads[0].name == "TEMPLATE"

    def test_reset_target_dir_fake_branch(self):
        """Try resetting target pipeline directory but original branch does not exist"""
        psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir)
        psync.inspect_sync_dir()
        psync.get_wf_config()

        psync.original_branch = "fake_branch"

        with pytest.raises(nf_core.pipelines.sync.SyncExceptionError) as exc_info:
            psync.reset_target_dir()
        assert exc_info.value.args[0].startswith("Could not reset to original branch `fake_branch`")

    def test_sync_no_changes(self):
        """Test pipeline sync when no changes are needed"""
        with mock.patch("requests.get", side_effect=mocked_requests_get), mock.patch(
            "requests.post", side_effect=mocked_requests_post
        ) as mock_post:
            psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir)

            # Mock that no changes were made
            psync.made_changes = False

            # Run sync
            psync.sync()

            # Verify no PR was created
            mock_post.assert_not_called()

    def test_sync_no_github_token(self):
        """Test sync fails appropriately when GitHub token is missing"""
        # Ensure GitHub token is not set
        if "GITHUB_AUTH_TOKEN" in os.environ:
            del os.environ["GITHUB_AUTH_TOKEN"]

        psync = nf_core.pipelines.sync.PipelineSync(self.pipeline_dir, make_pr=True)
        psync.made_changes = True  # Force changes to trigger PR attempt

        # Run sync and check for appropriate error
        with self.assertRaises(nf_core.pipelines.sync.PullRequestExceptionError) as exc_info:
            psync.sync()
        self.assertIn("GITHUB_AUTH_TOKEN not set!", str(exc_info.exception))



import logging
import os
from pathlib import Path
from typing import List, Optional, Union

import questionary
from rich import print
from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel
from rich.syntax import Syntax

import nf_core.components
import nf_core.modules.modules_utils
import nf_core.utils
from nf_core.components.components_command import ComponentCommand
from nf_core.components.components_utils import (
    NF_CORE_MODULES_NAME,
    get_components_to_install,
    prompt_component_version_sha,
)
from nf_core.modules.modules_json import ModulesJson

log = logging.getLogger(__name__)


class ComponentInstall(ComponentCommand):
    def __init__(
        self,
        pipeline_dir: Union[str, Path],
        component_type: str,
        force: bool = False,
        prompt: bool = False,
        sha: Optional[str] = None,
        remote_url: Optional[str] = None,
        branch: Optional[str] = None,
        no_pull: bool = False,
        installed_by: Optional[List[str]] = None,
    ):
        super().__init__(component_type, pipeline_dir, remote_url, branch, no_pull)
        self.force = force
        self.prompt = prompt
        self.sha = sha
        if installed_by is not None:
            self.installed_by = installed_by
        else:
            self.installed_by = [self.component_type]

    def install(self, component: str, silent: bool = False) -> bool:
        if self.repo_type == "modules":
            log.error(f"You cannot install a {component} in a clone of nf-core/modules")
            return False
        # Check whether pipelines is valid
        if not self.has_valid_directory():
            return False

        if self.component_type == "modules":
            # Check modules directory structure
            self.check_modules_structure()

        # Verify that 'modules.json' is consistent with the installed modules and subworkflows
        modules_json = ModulesJson(self.directory)
        if not silent:
            modules_json.check_up_to_date()

        # Verify that the remote repo's org_path does not match the org_path of any alternate repo among the installed modules
        if self.check_alternate_remotes(modules_json):
            err_msg = f"You are trying to install {self.component_type} from different repositories with the same organization name '{self.modules_repo.repo_path}' (set in the `.nf-core.yml` file in the `org_path` field).\nThis is not supported, and will likely cause problems. org_path should be set to the github account/organization name."
            log.error(err_msg)
            return False

        # Verify SHA
        if not self.modules_repo.verify_sha(self.prompt, self.sha):
            err_msg = f"SHA '{self.sha}' is not a valid commit SHA for the repository '{self.modules_repo.remote_url}'"
            log.error(err_msg)
            return False

        # verify self.modules_repo entries:
        if self.modules_repo is None:
            err_msg = "Could not find a valid modules repository."
            log.error(err_msg)
            return False
        if self.modules_repo.repo_path is None:
            err_msg = "Could not find a valid modules repository path."
            log.error(err_msg)
            return False

        # Check and verify component name

        component = self.collect_and_verify_name(component, self.modules_repo)
        if not component:
            return False

        # Get current version
        current_version = modules_json.get_component_version(
            self.component_type, component, self.modules_repo.remote_url, self.modules_repo.repo_path
        )

        # Set the install folder based on the repository name
        install_folder = Path(self.directory, self.component_type, self.modules_repo.repo_path)

        # Compute the component directory
        component_dir = Path(install_folder, component)

        # Check that the component is not already installed
        component_not_installed = self.check_component_installed(
            component, current_version, component_dir, self.modules_repo, self.force, self.prompt, silent
        )
        if not component_not_installed:
            log.debug(
                f"{self.component_type[:-1].title()} is already installed and force is not set.\nAdding the new installation source {self.installed_by} for {self.component_type[:-1]} {component} to 'modules.json' without installing the {self.component_type}."
            )
            modules_json.load()
            modules_json.update(self.component_type, self.modules_repo, component, current_version, self.installed_by)
            return False
        try:
            version = self.get_version(component, self.sha, self.prompt, current_version, self.modules_repo)
        except UserWarning as e:
            log.error(e)
            return False
        if not version:
            return False

        # Remove component if force is set and component is installed
        install_track = None
        if self.force:
            log.debug(f"Removing installed version of '{self.modules_repo.repo_path}/{component}'")
            self.clear_component_dir(component, component_dir)
            install_track = self.clean_modules_json(component, self.modules_repo, modules_json)
        if not silent:
            log.info(f"{'Rei' if self.force else 'I'}nstalling '{component}'")
        log.debug(
            f"Installing {self.component_type} '{component}' at modules hash {version} from {self.modules_repo.remote_url}"
        )

        # Download component files
        if not self.install_component_files(component, version, self.modules_repo, install_folder):
            return False

        # Update module.json with newly installed subworkflow
        modules_json.load()
        modules_json.update(
            self.component_type, self.modules_repo, component, version, self.installed_by, install_track
        )

        if self.component_type == "subworkflows":
            # Install included modules and subworkflows
            self.install_included_components(component_dir)

        if not silent:
            modules_json.load()
            modules_json.dump(run_prettier=True)
            # Print include statement
            component_name = "_".join(component.upper().split("/"))
            log.info(f"Use the following statement to include this {self.component_type[:-1]}:")
            Console().print(
                Syntax(
                    f"include {{ {component_name} }} from '../{Path(install_folder, component).relative_to(self.directory)}/main'",
                    "groovy",
                    theme="ansi_dark",
                    padding=1,
                )
            )
            if self.component_type == "subworkflows":
                subworkflow_config = Path(install_folder, component, "nextflow.config").relative_to(self.directory)
                if subworkflow_config.is_file():
                    log.info("Add the following config statement to use this subworkflow:")
                    Console().print(
                        Syntax(f"includeConfig '{subworkflow_config}'", "groovy", theme="ansi_dark", padding=1)
                    )
        return True

    def install_included_components(self, subworkflow_dir):
        """
        Install included modules and subworkflows
        """
        modules_to_install, subworkflows_to_install = get_components_to_install(subworkflow_dir)
        for s_install in subworkflows_to_install:
            original_installed = self.installed_by
            self.installed_by = [Path(subworkflow_dir).parts[-1]]
            self.install(s_install, silent=True)
            self.installed_by = original_installed
        for m_install in modules_to_install:
            original_component_type = self.component_type
            self.component_type = "modules"
            original_installed = self.installed_by
            self.installed_by = [Path(subworkflow_dir).parts[-1]]
            self.install(m_install, silent=True)
            self.component_type = original_component_type
            self.installed_by = original_installed

    def collect_and_verify_name(
        self, component: Optional[str], modules_repo: "nf_core.modules.modules_repo.ModulesRepo"
    ) -> str:
        """
        Collect component name.
        Check that the supplied name is an available module/subworkflow.
        """
        if component is None:
            component = questionary.autocomplete(
                f"{'Tool' if self.component_type == 'modules' else 'Subworkflow'} name:",
                choices=sorted(modules_repo.get_avail_components(self.component_type, commit=self.sha)),
                style=nf_core.utils.nfcore_question_style,
            ).unsafe_ask()

        if component is None:
            return ""

        # Check that the supplied name is an available module/subworkflow
        if component and component not in modules_repo.get_avail_components(self.component_type, commit=self.sha):
            log.error(f"{self.component_type[:-1].title()} '{component}' not found in available {self.component_type}")
            print(
                Panel(
                    Markdown(
                        f"Use the command `nf-core {self.component_type} list` to view available {self.component_type}."
                    ),
                    title="info",
                    title_align="left",
                    style="blue",
                    padding=1,
                )
            )

            raise ValueError

        if not modules_repo.component_exists(component, self.component_type, commit=self.sha):
            warn_msg = f"{self.component_type[:-1].title()} '{component}' not found in remote '{modules_repo.remote_url}' ({modules_repo.branch})"
            log.warning(warn_msg)

        return component

    def check_component_installed(self, component, current_version, component_dir, modules_repo, force, prompt, silent):
        """
        Check that the module/subworkflow is not already installed.

        Return:
            True: if the component is not installed
            False: if the component is installed
        """
        if (current_version is not None and os.path.exists(component_dir)) and not force:
            # make sure included components are also installed
            if self.component_type == "subworkflows":
                self.install_included_components(component_dir)
            if not silent:
                log.info(f"{self.component_type[:-1].title()} '{component}' is already installed.")

            if prompt:
                message = (
                    "?" if self.component_type == "modules" else " of this subworkflow and all it's imported modules?"
                )
                force = questionary.confirm(
                    f"{self.component_type[:-1].title()} {component} is already installed. \nDo you want to force the reinstallation{message}",
                    style=nf_core.utils.nfcore_question_style,
                    default=False,
                ).unsafe_ask()

            if not force:
                if not silent:
                    repo_flag = (
                        "" if modules_repo.repo_path == NF_CORE_MODULES_NAME else f"-g {modules_repo.remote_url} "
                    )
                    branch_flag = "" if modules_repo.branch == "master" else f"-b {modules_repo.branch} "

                    log.info(
                        f"To update '{component}' run 'nf-core {self.component_type} {repo_flag}{branch_flag}update {component}'. To force reinstallation use '--force'."
                    )
                return False

        return True

    def get_version(self, component, sha, prompt, current_version, modules_repo):
        """
        Get the version to install
        """
        if sha:
            version = sha
        elif prompt:
            try:
                version = prompt_component_version_sha(
                    component,
                    self.component_type,
                    installed_sha=current_version,
                    modules_repo=modules_repo,
                )
            except SystemError as e:
                log.error(e)
                return False
        else:
            # Fetch the latest commit for the module
            version = modules_repo.get_latest_component_version(component, self.component_type)
        return version

    def clean_modules_json(self, component, modules_repo, modules_json):
        """
        Remove installed version of module/subworkflow from modules.json
        """
        for repo_url, repo_content in modules_json.modules_json["repos"].items():
            for directory, dir_components in repo_content[self.component_type].items():
                for name, component_values in dir_components.items():
                    if name == component and directory == modules_repo.repo_path:
                        repo_to_remove = repo_url
                        log.debug(
                            f"Removing {self.component_type[:-1]} '{modules_repo.repo_path}/{component}' from repo '{repo_to_remove}' from modules.json."
                        )
                        modules_json.remove_entry(
                            self.component_type, component, repo_to_remove, modules_repo.repo_path
                        )
                        return component_values["installed_by"]

    def check_alternate_remotes(self, modules_json):
        """
        Check whether there are previously installed components with the same org_path but different remote urls
        Log error if multiple remotes exist.

        Return:
            True: if problematic components are found
            False: if problematic components are not found
        """
        modules_json.load()
        for repo_url, repo_content in modules_json.modules_json.get("repos", dict()).items():
            for component_type in repo_content:
                for directory in repo_content.get(component_type, dict()).keys():
                    if directory == self.modules_repo.repo_path and repo_url != self.modules_repo.remote_url:
                        return True
        return False



from pathlib import Path

import pytest

from nf_core.modules.modules_json import ModulesJson
from nf_core.subworkflows.install import SubworkflowInstall

from ..test_subworkflows import TestSubworkflows
from ..utils import (
    GITLAB_BRANCH_TEST_BRANCH,
    GITLAB_REPO,
    GITLAB_SUBWORKFLOWS_BRANCH,
    GITLAB_SUBWORKFLOWS_ORG_PATH_BRANCH,
    GITLAB_URL,
    with_temporary_folder,
)


class TestSubworkflowsInstall(TestSubworkflows):
    def test_subworkflows_install_bam_sort_stats_samtools(self):
        """Test installing a subworkflow - bam_sort_stats_samtools"""
        assert self.subworkflow_install.install("bam_sort_stats_samtools") is not False
        subworkflow_path = Path(
            self.subworkflow_install.directory, "subworkflows", "nf-core", "bam_sort_stats_samtools"
        )
        sub_subworkflow_path = Path(self.subworkflow_install.directory, "subworkflows", "nf-core", "bam_stats_samtools")
        samtools_index_path = Path(self.subworkflow_install.directory, "modules", "nf-core", "samtools", "index")
        samtools_sort_path = Path(self.subworkflow_install.directory, "modules", "nf-core", "samtools", "sort")
        samtools_stats_path = Path(self.subworkflow_install.directory, "modules", "nf-core", "samtools", "stats")
        samtools_idxstats_path = Path(self.subworkflow_install.directory, "modules", "nf-core", "samtools", "idxstats")
        samtools_flagstat_path = Path(self.subworkflow_install.directory, "modules", "nf-core", "samtools", "flagstat")
        assert subworkflow_path.exists()
        assert sub_subworkflow_path.exists()
        assert samtools_index_path.exists()
        assert samtools_sort_path.exists()
        assert samtools_stats_path.exists()
        assert samtools_idxstats_path.exists()
        assert samtools_flagstat_path.exists()

    def test_subworkflow_install_nopipeline(self):
        """Test installing a subworkflow - no pipeline given"""
        assert self.subworkflow_install.directory is not None
        self.subworkflow_install.directory = Path("non_existent_dir")
        assert self.subworkflow_install.install("bam_stats_samtools") is False

    @with_temporary_folder
    def test_subworkflows_install_emptypipeline(self, tmpdir):
        """Test installing a subworkflow - empty dir given"""

        Path(tmpdir, "nf-core-pipe").mkdir(exist_ok=True)
        self.subworkflow_install.directory = Path(tmpdir, "nf-core-pipe")
        with pytest.raises(UserWarning) as excinfo:
            self.subworkflow_install.install("bam_stats_samtools")
        assert "Could not find a 'main.nf' or 'nextflow.config' file" in str(excinfo.value)

    def test_subworkflows_install_nosubworkflow(self):
        """Test installing a subworkflow - unrecognised subworkflow given"""
        with pytest.raises(ValueError) as excinfo:
            self.subworkflow_install.install("foo")
        assert excinfo.typename == "ValueError"
        assert "Subworkflow 'foo' not found in available subworkflows" in self.caplog.text

    def test_subworkflows_install_bam_sort_stats_samtools_twice(self):
        """Test installing a subworkflow - bam_sort_stats_samtools already there"""
        self.subworkflow_install.install("bam_sort_stats_samtools")
        assert self.subworkflow_install.install("bam_sort_stats_samtools") is False

    def test_subworkflows_install_from_gitlab(self):
        """Test installing a subworkflow from GitLab"""
        assert self.subworkflow_install_gitlab.install("bam_stats_samtools") is True
        # Verify that the branch entry was added correctly
        modules_json = ModulesJson(self.pipeline_dir)
        assert (
            modules_json.get_component_branch(self.component_type, "bam_stats_samtools", GITLAB_URL, GITLAB_REPO)
            == GITLAB_SUBWORKFLOWS_BRANCH
        )

    def test_subworkflows_install_different_branch_fail(self):
        """Test installing a subworkflow from a different branch"""
        install_obj = SubworkflowInstall(self.pipeline_dir, remote_url=GITLAB_URL, branch=GITLAB_BRANCH_TEST_BRANCH)
        # The bam_stats_samtools subworkflow does not exists in the branch-test branch
        with pytest.raises(Exception) as excinfo:
            install_obj.install("bam_stats_samtools")
            assert "Subworkflow 'bam_stats_samtools' not found in available subworkflows" in str(excinfo.value)

    def test_subworkflows_install_tracking(self):
        """Test installing a subworkflow and finding the correct entries in installed_by section of modules.json"""
        assert self.subworkflow_install.install("bam_sort_stats_samtools")

        # Verify that the installed_by entry was added correctly
        modules_json = ModulesJson(self.pipeline_dir)
        mod_json = modules_json.get_modules_json()
        assert mod_json["repos"]["https://github.com/nf-core/modules.git"]["subworkflows"]["nf-core"][
            "bam_sort_stats_samtools"
        ]["installed_by"] == ["subworkflows"]
        assert mod_json["repos"]["https://github.com/nf-core/modules.git"]["subworkflows"]["nf-core"][
            "bam_stats_samtools"
        ]["installed_by"] == ["bam_sort_stats_samtools"]
        assert mod_json["repos"]["https://github.com/nf-core/modules.git"]["modules"]["nf-core"]["samtools/stats"][
            "installed_by"
        ] == ["bam_stats_samtools"]
        assert mod_json["repos"]["https://github.com/nf-core/modules.git"]["modules"]["nf-core"]["samtools/sort"][
            "installed_by"
        ] == ["bam_sort_stats_samtools"]

        # Clean directory
        self.subworkflow_remove.remove("bam_sort_stats_samtools")

    def test_subworkflows_install_tracking_added_already_installed(self):
        """Test installing a subworkflow and finding the correct entries in installed_by section of modules.json"""
        self.subworkflow_install.install("bam_sort_stats_samtools")
        self.subworkflow_install.install("bam_stats_samtools")

        # Verify that the installed_by entry was added correctly
        modules_json = ModulesJson(self.pipeline_dir)
        mod_json = modules_json.get_modules_json()
        assert mod_json["repos"]["https://github.com/nf-core/modules.git"]["subworkflows"]["nf-core"][
            "bam_sort_stats_samtools"
        ]["installed_by"] == ["subworkflows"]
        assert sorted(
            mod_json["repos"]["https://github.com/nf-core/modules.git"]["subworkflows"]["nf-core"][
                "bam_stats_samtools"
            ]["installed_by"]
        ) == sorted(["bam_sort_stats_samtools", "subworkflows"])

        # Clean directory
        self.subworkflow_remove.remove("bam_sort_stats_samtools")
        self.subworkflow_remove.remove("bam_stats_samtools")

    def test_subworkflows_install_tracking_added_super_subworkflow(self):
        """Test installing a subworkflow and finding the correct entries in installed_by section of modules.json"""
        self.subworkflow_install.install("bam_stats_samtools")
        self.subworkflow_install.install("bam_sort_stats_samtools")

        # Verify that the installed_by entry was added correctly
        modules_json = ModulesJson(self.pipeline_dir)
        mod_json = modules_json.get_modules_json()
        assert mod_json["repos"]["https://github.com/nf-core/modules.git"]["subworkflows"]["nf-core"][
            "bam_sort_stats_samtools"
        ]["installed_by"] == ["subworkflows"]
        assert sorted(
            mod_json["repos"]["https://github.com/nf-core/modules.git"]["subworkflows"]["nf-core"][
                "bam_stats_samtools"
            ]["installed_by"]
        ) == sorted(["subworkflows", "bam_sort_stats_samtools"])

    def test_subworkflows_install_alternate_remote(self):
        """Test installing a module from a different remote with the same organization path"""
        install_obj = SubworkflowInstall(
            self.pipeline_dir, remote_url=GITLAB_URL, branch=GITLAB_SUBWORKFLOWS_ORG_PATH_BRANCH
        )
        # Install a subworkflow from GitLab which is also installed from GitHub with the same org_path
        with pytest.raises(Exception) as excinfo:
            install_obj.install("fastqc")
            assert "Could not find a 'main.nf' or 'nextflow.config' file" in str(excinfo.value)
